import abc

from qibo.config import raise_error


class Backend(abc.ABC):
    def __init__(self):
        super().__init__()
        self.name = "backend"
        self.platform = None

        self.precision = "double"
        self.dtype = "complex128"
        self.matrices = None

        self.device = "/CPU:0"
        self.nthreads = 1
        self.supports_multigpu = False
        self.oom_error = MemoryError

    def __reduce__(self):
        """Allow pickling backend objects that have references to modules."""
        return self.__class__, tuple()

    def __repr__(self):
        if self.platform is None:
            return self.name
        else:
            return f"{self.name} ({self.platform})"

    @abc.abstractmethod
    def set_precision(self, precision):  # pragma: no cover
        """Set complex number precision.

        Args:
            precision (str): 'single' or 'double'.
        """
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def set_device(self, device):  # pragma: no cover
        """Set simulation device.

        Args:
            device (str): Device such as '/CPU:0', '/GPU:0', etc.
        """
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def set_threads(self, nthreads):  # pragma: no cover
        """Set number of threads for CPU simulation.

        Args:
            nthreads (int): Number of threads.
        """
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def cast(self, x, copy=False):  # pragma: no cover
        """Cast an object as the array type of the current backend.

        Args:
            x: Object to cast to array.
            copy (bool): If ``True`` a copy of the object is created in memory.
        """
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def issparse(self, x):  # pragma: no cover
        """Determine if a given array is a sparse tensor."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def to_numpy(self, x):  # pragma: no cover
        """Cast a given array to numpy."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def compile(self, func):  # pragma: no cover
        """Compile the given method.

        Available only for the tensorflow backend.
        """
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def zero_state(self, nqubits):  # pragma: no cover
        """Generate :math:`|000 \\cdots 0 \\rangle` state vector as an array."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def zero_density_matrix(self, nqubits):  # pragma: no cover
        """Generate :math:`|000\\cdots0\\rangle\\langle000\\cdots0|` density matrix as an array."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def identity_density_matrix(
        self, nqubits, normalize: bool = True
    ):  # pragma: no cover
        """Generate density matrix

        .. math::
            \\rho = \\frac{1}{2^\\text{nqubits}} \\, \\sum_{k=0}^{2^\\text{nqubits} - 1} \\,
                |k \\rangle \\langle k|

        if ``normalize=True``. If ``normalize=False``, returns the unnormalized
        Identity matrix, which is equivalent to :func:`numpy.eye`.
        """
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def plus_state(self, nqubits):  # pragma: no cover
        """Generate :math:`|+++\\cdots+\\rangle` state vector as an array."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def plus_density_matrix(self, nqubits):  # pragma: no cover
        """Generate :math:`|+++\\cdots+\\rangle\\langle+++\\cdots+|` density matrix as an array."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def matrix(self, gate):  # pragma: no cover
        """Convert a :class:`qibo.gates.Gate` to the corresponding matrix."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def matrix_parametrized(self, gate):  # pragma: no cover
        """Equivalent to :meth:`qibo.backends.abstract.Backend.matrix` for parametrized gates."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def matrix_fused(self, gate):  # pragma: no cover
        """Fuse matrices of multiple gates."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def apply_gate(self, gate, state, nqubits):  # pragma: no cover
        """Apply a gate to state vector."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def apply_gate_density_matrix(self, gate, state, nqubits):  # pragma: no cover
        """Apply a gate to density matrix."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def apply_gate_half_density_matrix(self, gate, state, nqubits):  # pragma: no cover
        """Apply a gate to one side of the density matrix."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def apply_channel(self, channel, state, nqubits):  # pragma: no cover
        """Apply a channel to state vector."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def apply_channel_density_matrix(self, channel, state, nqubits):  # pragma: no cover
        """Apply a channel to density matrix."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def collapse_state(
        self, state, qubits, shot, nqubits, normalize=True
    ):  # pragma: no cover
        """Collapse state vector according to measurement shot."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def collapse_density_matrix(
        self, state, qubits, shot, nqubits, normalize=True
    ):  # pragma: no cover
        """Collapse density matrix according to measurement shot."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def reset_error_density_matrix(self, gate, state, nqubits):  # pragma: no cover
        """Apply reset error to density matrix."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def thermal_error_density_matrix(self, gate, state, nqubits):  # pragma: no cover
        """Apply thermal relaxation error to density matrix."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def execute_circuit(
        self, circuit, initial_state=None, nshots=None
    ):  # pragma: no cover
        """Execute a :class:`qibo.models.circuit.Circuit`."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def execute_circuits(
        self, circuits, initial_states=None, nshots=None
    ):  # pragma: no cover
        """Execute multiple :class:`qibo.models.circuit.Circuit`s in parallel."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def execute_circuit_repeated(
        self, circuit, nshots, initial_state=None
    ):  # pragma: no cover
        """Execute a :class:`qibo.models.circuit.Circuit` multiple times.

        Useful for noise simulation using state vectors or for simulating gates
        controlled by measurement outcomes.
        """
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def execute_distributed_circuit(
        self, circuit, initial_state=None, nshots=None
    ):  # pragma: no cover
        """Execute a :class:`qibo.models.circuit.Circuit` using multiple GPUs."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def calculate_symbolic(
        self, state, nqubits, decimals=5, cutoff=1e-10, max_terms=20
    ):  # pragma: no cover
        """Dirac representation of a state vector."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def calculate_symbolic_density_matrix(
        self, state, nqubits, decimals=5, cutoff=1e-10, max_terms=20
    ):  # pragma: no cover
        """Dirac representation of a density matrix."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def calculate_probabilities(self, state, qubits, nqubits):  # pragma: no cover
        """Calculate probabilities given a state vector."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def calculate_probabilities_density_matrix(
        self, state, qubits, nqubits
    ):  # pragma: no cover
        """Calculate probabilities given a density matrix."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def set_seed(self, seed):  # pragma: no cover
        """Set the seed of the random number generator."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def sample_shots(self, probabilities, nshots):  # pragma: no cover
        """Sample measurement shots according to a probability distribution."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def aggregate_shots(self, shots):  # pragma: no cover
        """Collect shots to a single array."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def samples_to_binary(self, samples, nqubits):  # pragma: no cover
        """Convert samples from decimal representation to binary."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def samples_to_decimal(self, samples, nqubits):  # pragma: no cover
        """Convert samples from binary representation to decimal."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def calculate_frequencies(self, samples):  # pragma: no cover
        """Calculate measurement frequencies from shots."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def update_frequencies(
        self, frequencies, probabilities, nsamples
    ):  # pragma: no cover
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def sample_frequencies(self, probabilities, nshots):  # pragma: no cover
        """Sample measurement frequencies according to a probability distribution."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def partial_trace(self, state, qubits, nqubits):  # pragma: no cover
        """Trace out specific qubits of a state vector."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def partial_trace_density_matrix(self, state, qubits, nqubits):  # pragma: no cover
        """Trace out specific qubits of a density matrix."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def calculate_norm(self, state, order=2):  # pragma: no cover
        """Calculate norm of a state vector.

        For specifications on possible values of the parameter ``order``
        for the ``tensorflow`` backend, please refer to
        `tensorflow.norm <https://www.tensorflow.org/api_docs/python/tf/norm>`_.
        For all other backends, please refer to
        `numpy.linalg.norm <https://numpy.org/doc/stable/reference/generated/numpy.linalg.norm.html>`_.
        """
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def calculate_norm_density_matrix(self, state, order="nuc"):  # pragma: no cover
        """Calculate norm of a density matrix. Default is the ``nuclear`` norm.

        If ``order="nuc"``, it returns the nuclear norm of ``state``,
        assuming ``state`` is Hermitian (also known as trace norm).
        For specifications on the other  possible values of the
        parameter ``order`` for the ``tensorflow`` backend, please refer to
        `tensorflow.norm <https://www.tensorflow.org/api_docs/python/tf/norm>`_.
        For all other backends, please refer to
        `numpy.linalg.norm <https://numpy.org/doc/stable/reference/generated/numpy.linalg.norm.html>`_.
        """
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def calculate_overlap(self, state1, state2):  # pragma: no cover
        """Calculate overlap of two state vectors."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def calculate_overlap_density_matrix(self, state1, state2):  # pragma: no cover
        """Calculate overlap of two density matrices."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def calculate_eigenvalues(self, matrix, k=6):  # pragma: no cover
        """Calculate eigenvalues of a matrix."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def calculate_eigenvectors(self, matrix, k=6):  # pragma: no cover
        """Calculate eigenvectors of a matrix."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def calculate_matrix_exp(
        self, matrix, a, eigenvectors=None, eigenvalues=None
    ):  # pragma: no cover
        """Calculate matrix exponential of a matrix.

        If the eigenvectors and eigenvalues are given the matrix diagonalization is
        used for exponentiation.
        """
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def calculate_expectation_state(
        self, hamiltonian, state, normalize
    ):  # pragma: no cover
        """Calculate expectation value of a state vector given the observable matrix."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def calculate_expectation_density_matrix(
        self, hamiltonian, state, normalize
    ):  # pragma: no cover
        """Calculate expectation value of a density matrix given the observable matrix."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def calculate_hamiltonian_matrix_product(
        self, matrix1, matrix2
    ):  # pragma: no cover
        """Multiply two matrices."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def calculate_hamiltonian_state_product(self, matrix, state):  # pragma: no cover
        """Multiply a matrix to a state vector or density matrix."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def assert_allclose(self, value, target, rtol=1e-7, atol=0.0):  # pragma: no cover
        raise_error(NotImplementedError)

    def assert_circuitclose(self, circuit, target_circuit, rtol=1e-7, atol=0.0):
        value = self.execute_circuit(circuit)._state
        target = self.execute_circuit(target_circuit)._state
        self.assert_allclose(value, target, rtol=rtol, atol=atol)

    @abc.abstractmethod
    def test_regressions(self, name):  # pragma: no cover
        """Correct outcomes for tests that involve random numbers.

        The outcomes of such tests depend on the backend.
        """
        raise_error(NotImplementedError)



"""Module definig the Clifford object, which allows phase-space representation of Clifford circuits and stabilizer states."""

from dataclasses import dataclass, field
from functools import reduce
from itertools import product
from typing import Optional, Union

import numpy as np

from qibo import Circuit
from qibo.backends import CliffordBackend
from qibo.config import raise_error
from qibo.gates import M
from qibo.measurements import frequencies_to_binary

from ._clifford_utils import _decomposition_AG04, _decomposition_BM20, _string_product


@dataclass
class Clifford:
    """Object storing the results of a circuit execution with the :class:`qibo.backends.clifford.CliffordBackend`.

    Args:
        data (ndarray or :class:`qibo.models.circuit.Circuit`): If ``ndarray``, it is the
            symplectic matrix of the stabilizer state in phase-space representation.
            If :class:`qibo.models.circuit.Circuit`, it is a circuit composed only of Clifford
            gates and computational-basis measurements.
        nqubits (int, optional): number of qubits of the state.
        measurements (list, optional): list of measurements gates :class:`qibo.gates.M`.
            Defaults to ``None``.
        nshots (int, optional): number of shots used for sampling the measurements.
            Defaults to :math:`1000`.
        engine (str, optional): engine to use in the execution of the
            :class:`qibo.backends.CliffordBackend`. It accepts ``"numpy"``, ``"numba"``,
            ``"cupy"``, and ``"stim"`` (see `stim <https://github.com/quantumlib/Stim>`_).
            If ``None``, defaults to the corresponding engine
            from :class:`qibo.backends.GlobalBackend`. Defaults to ``None``.
    """

    symplectic_matrix: np.ndarray = field(init=False)
    data: Union[np.ndarray, Circuit] = field(repr=False)
    nqubits: Optional[int] = None
    measurements: Optional[list] = None
    nshots: int = 1000
    engine: Optional[str] = None

    _backend: Optional[CliffordBackend] = None
    _measurement_gate = None
    _samples: Optional[int] = None

    def __post_init__(self):
        if isinstance(self.data, Circuit):
            clifford = self.from_circuit(self.data, engine=self.engine)
            self.symplectic_matrix = clifford.symplectic_matrix
            self.nqubits = clifford.nqubits
            self.measurements = clifford.measurements
            self._samples = clifford._samples
            self._measurement_gate = clifford._measurement_gate
        else:
            # adding the scratch row if not provided
            self.symplectic_matrix = self.data
            if self.symplectic_matrix.shape[0] % 2 == 0:
                self.symplectic_matrix = np.vstack(
                    (self.symplectic_matrix, np.zeros(self.symplectic_matrix.shape[1]))
                )
            self.nqubits = int((self.symplectic_matrix.shape[1] - 1) / 2)
        if self._backend is None:
            self._backend = CliffordBackend(self.engine)
        self.engine = self._backend.engine

    @classmethod
    def from_circuit(
        cls,
        circuit: Circuit,
        initial_state: Optional[np.ndarray] = None,
        nshots: int = 1000,
        engine: Optional[str] = None,
    ):
        """Allows to create a :class:`qibo.quantum_info.clifford.Clifford` object by executing the input circuit.

        Args:
            circuit (:class:`qibo.models.circuit.Circuit`): Clifford circuit to run.
            initial_state (ndarray, optional): symplectic matrix of the initial state.
                If ``None``, defaults to the symplectic matrix of the zero state.
                Defaults to ``None``.
            nshots (int, optional): number of measurement shots to perform
                if ``circuit`` has measurement gates. Defaults to :math:`10^{3}`.
            engine (str, optional): engine to use in the execution of the
                :class:`qibo.backends.CliffordBackend`. It accepts ``"numpy"``, ``"numba"``,
                ``"cupy"``, and ``"stim"`` (see `stim <https://github.com/quantumlib/Stim>`_).
                If ``None``, defaults to the corresponding engine
                from :class:`qibo.backends.GlobalBackend`. Defaults to ``None``.

        Returns:
            (:class:`qibo.quantum_info.clifford.Clifford`): Object storing the result of the circuit execution.
        """
        cls._backend = CliffordBackend(engine)

        return cls._backend.execute_circuit(circuit, initial_state, nshots)

    def to_circuit(self, algorithm: Optional[str] = "AG04"):
        """Converts symplectic matrix into a Clifford circuit.

        Args:
            algorithm (str, optional): If ``AG04``, uses the decomposition algorithm from
                `Aaronson & Gottesman (2004) <https://arxiv.org/abs/quant-ph/0406196>`_.
                If ``BM20`` and ``Clifford.nqubits <= 3``, uses the decomposition algorithm from
                `Bravyi & Maslov (2020) <https://arxiv.org/abs/2003.09412>`_.
                Defaults to ``AG04``.

        Returns:
            :class:`qibo.models.circuit.Circuit`: circuit composed of Clifford gates.
        """
        if not isinstance(algorithm, str):
            raise_error(
                TypeError,
                f"``algorithm`` must be type str, but it is type {type(algorithm)}",
            )

        if algorithm not in ["AG04", "BM20"]:
            raise_error(ValueError, f"``algorithm`` {algorithm} not found.")

        if algorithm == "BM20":
            return _decomposition_BM20(self)

        return _decomposition_AG04(self)

    def generators(self, return_array: bool = False):
        """Extracts the generators of stabilizers and destabilizers.

        Args:
            return_array (bool, optional): If ``True`` returns the generators as ``ndarray``.
                If ``False``, their representation as strings is returned. Defaults to ``False``.

        Returns:
            (list, list): Generators and their corresponding phases, respectively.
        """
        return self._backend.symplectic_matrix_to_generators(
            self.symplectic_matrix, return_array
        )

    def stabilizers(self, symplectic: bool = False, return_array: bool = False):
        """Extracts the stabilizers of the state.

        Args:
            symplectic (bool, optional): If ``True``, returns the rows of the symplectic matrix
                that correspond to the :math:`n` generators of the :math:`2^{n}` total stabilizers,
                independently of ``return_array``.
            return_array (bool, optional): To be used when ``symplectic = False``.
                If ``True`` returns the stabilizers as ``ndarray``.
                If ``False``, returns stabilizers as strings. Defaults to ``False``.

        Returns:
            (ndarray or list): Stabilizers of the state.
        """
        if not symplectic:
            generators, phases = self.generators(return_array)

            return self._construct_operators(
                generators[self.nqubits :],
                phases[self.nqubits :],
            )

        return self.symplectic_matrix[self.nqubits : -1, :]

    def destabilizers(self, symplectic: bool = False, return_array: bool = False):
        """Extracts the destabilizers of the state.

        Args:
            symplectic (bool, optional): If ``True``, returns the rows of the symplectic matrix
                that correspond to the :math:`n` generators of the :math:`2^{n}` total
                destabilizers, independently of ``return_array``.
            return_array (bool, optional): To be used when ``symplectic = False``.
                If ``True`` returns the destabilizers as ``ndarray``.
                If ``False``, their representation as strings is returned.
                Defaults to ``False``.

        Returns:
            (ndarray or list): Destabilizers of the state.
        """
        if not symplectic:
            generators, phases = self.generators(return_array)

            return self._construct_operators(
                generators[: self.nqubits], phases[: self.nqubits]
            )

        return self.symplectic_matrix[: self.nqubits, :]

    def state(self):
        """Builds the density matrix representation of the state.

        .. note::
            This method is inefficient in runtime and memory for a large number of qubits.

        Returns:
            (ndarray): Density matrix of the state.
        """
        stabilizers = self.stabilizers(return_array=True)

        return self.engine.np.sum(stabilizers, axis=0) / len(stabilizers)

    @property
    def measurement_gate(self):
        """Single measurement gate containing all measured qubits.

        Useful for sampling all measured qubits at once when simulating.
        """
        if self._measurement_gate is None:
            for gate in self.measurements:
                if self._measurement_gate is None:
                    self._measurement_gate = M(*gate.init_args, **gate.init_kwargs)
                else:
                    self._measurement_gate.add(gate)

        return self._measurement_gate

    def samples(self, binary: bool = True, registers: bool = False):
        """Returns raw measurement samples.

        Args:
            binary (bool, optional): If ``False``, return samples in binary form.
                If ``True``, returns samples in decimal form. Defalts to ``True``.
            registers (bool, optional): If ``True``, groups samples according to registers.
                Defaults to ``False``.

        Returns:
            If ``binary`` is ``True``
                samples are returned in binary form as a tensor
                of shape ``(nshots, n_measured_qubits)``.
            If ``binary`` is ``False``
                samples are returned in decimal form as a tensor
                of shape ``(nshots,)``.
            If ``registers`` is ``True``
                samples are returned in a ``dict`` where the keys are the register
                names and the values are the samples tensors for each register.
            If ``registers`` is ``False``
                a single tensor is returned which contains samples from all the
                measured qubits, independently of their registers.
        """
        if not self.measurements:
            raise_error(RuntimeError, "No measurement provided.")

        measured_qubits = self.measurement_gate.qubits

        if self._samples is None:
            if self.measurements[0].result.has_samples():
                samples = np.concatenate(
                    [gate.result.samples() for gate in self.measurements], axis=1
                )
            else:
                samples = self._backend.sample_shots(
                    self.symplectic_matrix, measured_qubits, self.nqubits, self.nshots
                )
            if self.measurement_gate.has_bitflip_noise():
                p0, p1 = self.measurement_gate.bitflip_map
                bitflip_probabilities = self._backend.cast(
                    [
                        [p0.get(q) for q in measured_qubits],
                        [p1.get(q) for q in measured_qubits],
                    ]
                )
                samples = self._backend.cast(samples, dtype="int32")
                samples = self._backend.apply_bitflips(samples, bitflip_probabilities)
            # register samples to individual gate ``MeasurementResult``
            qubit_map = {
                q: i for i, q in enumerate(self.measurement_gate.target_qubits)
            }
            self._samples = self._backend.cast(samples, dtype="int32")
            for gate in self.measurements:
                rqubits = tuple(qubit_map.get(q) for q in gate.target_qubits)
                gate.result.register_samples(self._samples[:, rqubits], self._backend)

        if registers:
            return {
                gate.register_name: gate.result.samples(binary)
                for gate in self.measurements
            }

        if binary:
            return self._samples

        return self._backend.samples_to_decimal(self._samples, len(measured_qubits))

    def frequencies(self, binary: bool = True, registers: bool = False):
        """Returns the frequencies of measured samples.

        Args:
            binary (bool, optional): If ``True``, return frequency keys in binary form.
                If ``False``, return frequency keys in decimal form. Defaults to ``True``.
            registers (bool, optional): If ``True``, groups frequencies according to registers.
                Defaults to ``False``.

        Returns:
            A `collections.Counter` where the keys are the observed values
            and the values the corresponding frequencies, that is the number
            of times each measured value/bitstring appears.

            If ``binary`` is ``True``
                the keys of the `Counter` are in binary form, as strings of
                :math:`0`s and :math`1`s.
            If ``binary`` is ``False``
                the keys of the ``Counter`` are integers.
            If ``registers`` is ``True``
                a `dict` of `Counter` s is returned where keys are the name of
                each register.
            If ``registers`` is ``False``
                a single ``Counter`` is returned which contains samples from all
                the measured qubits, independently of their registers.
        """
        measured_qubits = self.measurement_gate.target_qubits
        freq = self._backend.calculate_frequencies(self.samples(False))
        if registers:
            if binary:
                return {
                    gate.register_name: frequencies_to_binary(
                        self._backend.calculate_frequencies(gate.result.samples(False)),
                        len(gate.target_qubits),
                    )
                    for gate in self.measurements
                }

            return {
                gate.register_name: self._backend.calculate_frequencies(
                    gate.result.samples(False)
                )
                for gate in self.measurements
            }

        if binary:
            return frequencies_to_binary(freq, len(measured_qubits))

        return freq

    def probabilities(self, qubits: Optional[Union[tuple, list]] = None):
        """Computes the probabilities of the selected qubits from the measured samples.

        Args:
            qubits (tuple or list, optional): Qubits for which to compute the probabilities.

        Returns:
            (ndarray): Measured probabilities.
        """
        if isinstance(qubits, list):
            qubits = tuple(qubits)

        measured_qubits = self.measurement_gate.qubits
        if qubits is not None:
            if not set(qubits).issubset(set(measured_qubits)):
                raise_error(
                    RuntimeError,
                    f"Asking probabilities for qubits {qubits}, but only qubits {measured_qubits} were measured.",
                )
            qubits = [measured_qubits.index(q) for q in qubits]
        else:
            qubits = range(len(measured_qubits))

        probs = [0 for _ in range(2 ** len(measured_qubits))]
        samples = self.samples(binary=False)
        for s in samples:
            probs[int(s)] += 1

        probs = self.engine.cast(probs, float) / len(samples)

        return self._backend.calculate_probabilities(
            self.engine.np.sqrt(probs), qubits, len(measured_qubits)
        )

    def copy(self, deep: bool = False):
        """Returns copy of :class:`qibo.quantum_info.clifford.Clifford` object.

        Args:
            deep (bool, optional): If ``True``, creates another copy in memory.
                Defaults to ``False``.

        Returns:
            :class:`qibo.quantum_info.clifford.Clifford`: copy of original ``Clifford`` object.
        """
        if not isinstance(deep, bool):
            raise_error(
                TypeError, f"``deep`` must be type bool, but it is type {type(deep)}."
            )

        symplectic_matrix = (
            self.engine.np.copy(self.symplectic_matrix)
            if deep
            else self.symplectic_matrix
        )

        return self.__class__(
            symplectic_matrix,
            self.nqubits,
            self.measurements,
            self.nshots,
            _backend=self._backend,
        )

    def _construct_operators(self, generators: list, phases: list):
        """Helper function to construct all the operators from their generators.

        Args:
            generators (list or ndarray): generators.
            phases (list or ndarray): phases of the generators.

        Returns:
            (list): All operators generated by the generators of the stabilizer group.
        """

        if not isinstance(generators[0], str):
            generators = self._backend.cast(generators)
            phases = self._backend.cast(phases)

            operators = generators * phases.reshape(-1, 1, 1)
            identity = self.engine.identity_density_matrix(
                self.nqubits, normalize=False
            )
            operators = self._backend.cast([(g, identity) for g in operators])

            return self._backend.cast(
                [reduce(self.engine.np.matmul, ops) for ops in product(*operators)]
            )

        operators = list(np.copy(generators))
        for i in (phases == -1).nonzero()[0]:
            i = int(i)
            operators[i] = f"-{operators[i]}"

        identity = "".join(["I" for _ in range(self.nqubits)])

        operators = [(g, identity) for g in operators]

        return [_string_product(ops) for ops in product(*operators)]



import numpy as np

from qibo import gates
from qibo.backends import matrices
from qibo.config import raise_error
from qibo.hamiltonians import Hamiltonian
from qibo.models.circuit import Circuit


class qPDF:
    """Variational Circuit for Quantum PDFs (qPDF).

    Args:
        ansatz (str): the ansatz name, options are 'Weighted' and 'Fourier'.
        layers (int): the number of layers for the ansatz.
        nqubits (int): the number of qubits for the circuit.
        multi_output (bool): allocates a multi-output model per PDF flavour (default is False).
        backend (:class:`qibo.backends.abstract.Backend`): Backend object to use for execution.
            If ``None`` the currently active global backend is used.
            Default is ``None``.
    """

    def __init__(self, ansatz, layers, nqubits, multi_output=False, backend=None):
        """Initialize qPDF."""
        if not isinstance(layers, int) or layers < 1:  # pragma: no cover
            raise_error(RuntimeError, "Layers must be positive and integer.")
        if not isinstance(nqubits, int) or nqubits < 1:  # pragma: no cover
            raise_error(RuntimeError, "Number of qubits must be positive and integer.")
        if not isinstance(multi_output, bool):  # pragma: no cover
            raise_error(TypeError, "multi-output must be a boolean.")

        # parse ansatz
        if ansatz == "Weighted":
            ansatz_function = ansatz_Weighted
        elif ansatz == "Fourier":
            ansatz_function = ansatz_Fourier
        else:  # pragma: no cover
            raise_error(NotImplementedError, f"Ansatz {ansatz} not found.")

        # load ansatz
        self.circuit, self.rotation, self.nparams = ansatz_function(layers, nqubits)

        # load backend
        from qibo.backends import _check_backend

        self.backend = _check_backend(backend)

        # load hamiltonian
        if multi_output:
            self.hamiltonian = [
                qpdf_hamiltonian(nqubits, z_qubit=q, backend=self.backend)
                for q in range(nqubits)
            ]
        else:
            self.hamiltonian = [qpdf_hamiltonian(nqubits, backend=self.backend)]

    def _model(self, state, hamiltonian):
        """Internal function for the evaluation of PDFs.

        Args:
            state (numpy.array): state vector.
            hamiltonian (qibo.hamiltonian.Hamiltonian): the Hamiltonian object.

        Returs:
            The qPDF object following the (1-z)/(1+z) structure.
        """
        z = hamiltonian.expectation(state)
        y = (1 - z) / (1 + z)
        return y

    def predict(self, parameters, x):
        """Predict PDF model from underlying circuit.

        Args:
            parameters (numpy.array): list of parameters for the gates.
            x (numpy.array): a numpy array with the points in x to be evaluated.

        Returns:
            A numpy array with the PDF values.
        """
        if len(parameters) != self.nparams:  # pragma: no cover
            raise_error(
                RuntimeError, "Mismatch between number of parameters and model size."
            )
        pdf = np.zeros(shape=(len(x), len(self.hamiltonian)))
        for i, x_value in enumerate(x):
            params = self.rotation(parameters, x_value)
            self.circuit.set_parameters(params)
            result = self.backend.execute_circuit(self.circuit)
            state = result.state()
            for flavour, flavour_hamiltonian in enumerate(self.hamiltonian):
                pdf[i, flavour] = self._model(state, flavour_hamiltonian)
        return pdf


def qpdf_hamiltonian(nqubits, z_qubit=0, backend=None):
    """Precomputes Hamiltonian.

    Args:
        nqubits (int): number of qubits.
        z_qubit (int): qubit where the Z measurement is applied, must be z_qubit < nqubits
        backend (:class:`qibo.backends.abstract.Backend`): Backend object to use for execution.
            If ``None`` the currently active global backend is used.
            Default is ``None``.

    Returns:
        An Hamiltonian object.
    """
    eye = matrices.I
    if z_qubit == 0:
        h = matrices.Z
        for _ in range(nqubits - 1):
            h = np.kron(eye, h)

    elif z_qubit == nqubits - 1:
        h = eye
        for _ in range(nqubits - 2):
            h = np.kron(eye, h)
        h = np.kron(matrices.Z, h)
    else:
        h = eye
        for _ in range(nqubits - 1):
            if _ + 1 == z_qubit:
                h = np.kron(matrices.Z, h)
            else:
                h = np.kron(eye, h)
    return Hamiltonian(nqubits, h, backend=backend)


def map_to(x):
    """Auxiliary function"""
    return 2 * np.pi * x


def maplog_to(x):
    """Auxiliary function"""
    return -np.pi * np.log10(x)


def ansatz_Fourier(layers, qubits=1):
    """Fourier Ansatz implementation. It is composed by 3 parameters per layer
    and qubit: U3(a, b, c) Ry(x) || U3(a, b, c) Ry(log x).

    Args:
        layers (int): number of layers.
        qubits (int): number of qubits.

    Returns:
        The circuit, the rotation function and the total number of parameters.
    """
    circuit = Circuit(qubits)
    for l in range(layers - 1):
        for q in range(qubits):
            for _ in range(2):
                circuit.add(gates.RY(q, theta=0))
                circuit.add(gates.RZ(q, theta=0))
                circuit.add(gates.RY(q, theta=0))

        if qubits > 1:
            for q in range(0, qubits, 2):
                circuit.add(gates.CU1(q, q + 1, theta=0))
        if qubits > 2:
            for q in range(1, qubits + 1, 2):
                circuit.add(gates.CU1(q, (q + 1) % qubits, theta=0))

    for q in range(qubits):
        for _ in range(2):
            circuit.add(gates.RY(q, theta=0))
            circuit.add(gates.RZ(q, theta=0))
            circuit.add(gates.RY(q, theta=0))

    def rotation(theta, x):
        p = circuit.get_parameters()
        i = 0
        j = 0
        for l in range(layers - 1):
            for q in range(qubits):
                p[i] = map_to(x)
                p[i + 1 : i + 3] = theta[j : j + 2]
                i += 3
                j += 2

                p[i] = 0.5 * maplog_to(x)
                p[i + 1 : i + 3] = theta[j : j + 2]
                i += 3
                j += 2
            if qubits > 1:
                for q in range(0, qubits, 2):
                    p[i] = theta[j]
                    i += 1
                    j += 1
            if qubits > 2:
                for q in range(1, qubits + 1, 2):
                    p[i] = theta[j]
                    i += 1
                    j += 1
        for q in range(qubits):
            p[i] = 0.5 * map_to(x)
            p[i + 1 : i + 3] = theta[j : j + 2]
            i += 3
            j += 2

            p[i] = 0.5 * maplog_to(x)
            p[i + 1 : i + 3] = theta[j : j + 2]
            i += 3
            j += 2
        return p

    nparams = 4 * layers * qubits + (layers - 1) * int(np.ceil(qubits / 2)) * (
        int(qubits > 1) + int(qubits > 2)
    )

    return circuit, rotation, nparams


def ansatz_Weighted(layers, qubits=1):
    """Fourier Ansatz implementation. 4 parameters per layer and
    qubit: Ry(wx + a), Rz(v log(x) + b)

    Args:
        layers (int): number of layers.
        qubits (int): number of qubits.

    Returns:
        The circuit, the rotation function and the total number of
        parameters.
    """
    circuit = Circuit(qubits)
    for _ in range(layers - 1):
        for q in range(qubits):
            circuit.add(gates.RY(q, theta=0))
            circuit.add(gates.RZ(q, theta=0))
        if qubits > 1:
            for q in range(0, qubits, 2):
                circuit.add(gates.CU1(q, (q + 1) % qubits, theta=0))
        if qubits > 2:
            for q in range(1, qubits + 1, 2):
                circuit.add(gates.CU1(q, (q + 1) % qubits, theta=0))
    for q in range(qubits):
        circuit.add(gates.RY(q, theta=0))
        circuit.add(gates.RZ(q, theta=0))

    def rotation(theta, x):
        p = circuit.get_parameters()
        i = 0
        j = 0
        for l in range(layers - 1):
            for q in range(qubits):
                p[i] = theta[j] + theta[j + 1] * map_to(x)
                p[i + 1] = theta[j + 2] + theta[j + 3] * maplog_to(x)
                i += 2
                j += 4

            if qubits > 1:
                for q in range(0, qubits, 2):
                    p[i] = theta[j]
                    i += 1
                    j += 1
            if qubits > 2:
                for q in range(1, qubits + 1, 2):
                    p[i] = theta[j]
                    i += 1
                    j += 1

        for q in range(qubits):
            p[i] = theta[j] + theta[j + 1] * map_to(x)
            p[i + 1] = theta[j + 2] + theta[j + 3] * maplog_to(x)
            i += 2
            j += 4

        return p

    nparams = 4 * layers * qubits + (layers - 1) * int(np.ceil(qubits / 2)) * (
        int(qubits > 1) + int(qubits > 2)
    )
    return circuit, rotation, nparams



import numpy as np

from qibo import gates, matrices
from qibo.backends import NumpyBackend, _check_backend
from qibo.config import raise_error

magic_basis = np.array(
    [[1, -1j, 0, 0], [0, 0, 1, -1j], [0, 0, -1, -1j], [1, 1j, 0, 0]]
) / np.sqrt(2)

bell_basis = np.array(
    [[1, 1, 0, 0], [0, 0, 1, 1], [0, 0, 1, -1], [1, -1, 0, 0]]
) / np.sqrt(2)

H = np.array([[1, 1], [1, -1]]) / np.sqrt(2)


def u3_decomposition(unitary):
    """Decomposes arbitrary one-qubit gates to U3.

    Args:
        unitary (ndarray): Unitary :math:`2 \\times 2` matrix to be decomposed.

    Returns:
        (float, float, float): parameters of U3 gate.
    """
    # https://github.com/Qiskit/qiskit-terra/blob/d2e3340adb79719f9154b665e8f6d8dc26b3e0aa/qiskit/quantum_info/synthesis/one_qubit_decompose.py#L221
    su2 = unitary / np.sqrt(np.linalg.det(unitary))
    theta = 2 * np.arctan2(abs(su2[1, 0]), abs(su2[0, 0]))
    plus = np.angle(su2[1, 1])
    minus = np.angle(su2[1, 0])
    phi = plus + minus
    lam = plus - minus

    return theta, phi, lam


def calculate_psi(unitary, magic_basis=magic_basis, backend=None):
    """Solves the eigenvalue problem of :math:`U^{T} U`.

    See step (1) of Appendix A in arXiv:quant-ph/0011050.

    Args:
        unitary (ndarray): Unitary matrix of the gate we are decomposing
            in the computational basis.
        magic_basis (ndarray, optional): basis in which to solve the eigenvalue problem.
            Defaults to ``magic basis``.
        backend (:class:`qibo.backends.abstract.Backend`): Backend to use for calculations.

    Returns:
        (ndarray) Eigenvectors in the computational basis and eigenvalues of :math:`U^{T} U`.
    """
    backend = _check_backend(backend)

    if backend.__class__.__name__ in [
        "CupyBackend",
        "CuQuantumBackend",
    ]:  # pragma: no cover
        raise_error(
            NotImplementedError,
            f"{backend.__class__.__name__} does not support `linalg.eig.`",
        )

    magic_basis = backend.cast(magic_basis)
    unitary = backend.cast(unitary)
    # write unitary in magic basis
    u_magic = (
        backend.np.transpose(backend.np.conj(magic_basis), (1, 0))
        @ unitary
        @ magic_basis
    )
    # construct and diagonalize UT_U
    ut_u = backend.np.transpose(u_magic, (1, 0)) @ u_magic
    # When the matrix given to np.linalg.eig is a diagonal matrix up to machine precision the decomposition
    # is not accurate anymore. decimals = 20 works for random 2q Clifford unitaries.
    if backend.__class__.__name__ == "TensorflowBackend":
        eigvals, psi_magic = np.linalg.eig(np.round(ut_u, decimals=20))
        psi_magic, _ = np.linalg.qr(psi_magic)
    else:
        eigvals, psi_magic = backend.np.linalg.eig(np.round(ut_u, decimals=20))
        # orthogonalize eigenvectors in the case of degeneracy (Gram-Schmidt)
        psi_magic, _ = backend.np.linalg.qr(psi_magic)
    # write psi in computational basis
    psi = backend.np.matmul(magic_basis, psi_magic)
    return psi, eigvals


def schmidt_decompose(state, backend=None):
    """Decomposes a two-qubit product state to its single-qubit parts.

    Args:
        state (ndarray): product state to be decomposed.

    Returns:
        (ndarray, ndarray): decomposition

    """
    backend = _check_backend(backend)
    if backend.__class__.__name__ == "TensorflowBackend":
        u, d, v = np.linalg.svd(backend.np.reshape(state, (2, 2)))
    else:
        u, d, v = backend.np.linalg.svd(backend.np.reshape(state, (2, 2)))
    if not np.allclose(d, [1, 0]):  # pragma: no cover
        raise_error(
            ValueError,
            f"Unexpected singular values: {d}\nCan only decompose product states.",
        )
    return u[:, 0], v[0]


def calculate_single_qubit_unitaries(psi, backend=None):
    """Calculates local unitaries that maps a maximally entangled basis to the magic basis.

    See Lemma 1 of Appendix A in arXiv:quant-ph/0011050.

    Args:
        psi (ndarray): Maximally entangled two-qubit states that define a basis.

    Returns:
        (ndarray, ndarray): Local unitaries UA and UB that map the given basis to the magic basis.
    """
    backend = _check_backend(backend)
    psi_magic = backend.np.matmul(backend.np.conj(backend.cast(magic_basis)).T, psi)
    if not np.allclose(
        backend.to_numpy(psi_magic).imag, np.zeros_like(psi_magic)
    ):  # pragma: no cover
        raise_error(NotImplementedError, "Given state is not real in the magic basis.")
    psi_bar = backend.cast(psi.T, copy=True)

    # find e and f by inverting (A3), (A4)
    ef = (psi_bar[0] + 1j * psi_bar[1]) / np.sqrt(2)
    e_f_ = (psi_bar[0] - 1j * psi_bar[1]) / np.sqrt(2)
    e, f = schmidt_decompose(ef, backend=backend)
    e_, f_ = schmidt_decompose(e_f_, backend=backend)
    # find exp(1j * delta) using (A5a)
    ef_ = backend.np.kron(e, f_)
    phase = (
        1j
        * np.sqrt(2)
        * backend.np.sum(backend.np.multiply(backend.np.conj(ef_), psi_bar[2]))
    )
    v0 = backend.cast(np.asarray([1, 0]))
    v1 = backend.cast(np.asarray([0, 1]))
    # construct unitaries UA, UB using (A6a), (A6b)
    ua = backend.np.tensordot(v0, backend.np.conj(e), 0) + phase * backend.np.tensordot(
        v1, backend.np.conj(e_), 0
    )
    ub = backend.np.tensordot(v0, backend.np.conj(f), 0) + backend.np.conj(
        phase
    ) * backend.np.tensordot(v1, backend.np.conj(f_), 0)
    return ua, ub


def calculate_diagonal(unitary, ua, ub, va, vb, backend=None):
    """Calculates Ud matrix that can be written as exp(-iH).

    See Eq. (A1) in arXiv:quant-ph/0011050.
    Ud is diagonal in the magic and Bell basis.
    """
    backend = _check_backend(backend)
    # normalize U_A, U_B, V_A, V_B so that detU_d = 1
    # this is required so that sum(lambdas) = 0
    # and Ud can be written as exp(-iH)
    if backend.__class__.__name__ == "TensorflowBackend":
        det = np.linalg.det(unitary) ** (1 / 16)
    else:
        det = backend.np.linalg.det(unitary) ** (1 / 16)
    ua *= det
    ub *= det
    va *= det
    vb *= det
    u_dagger = backend.np.transpose(
        backend.np.conj(
            backend.np.kron(
                ua,
                ub,
            )
        ),
        (1, 0),
    )
    v_dagger = backend.np.transpose(backend.np.conj(backend.np.kron(va, vb)), (1, 0))
    ud = u_dagger @ unitary @ v_dagger
    return ua, ub, ud, va, vb


def magic_decomposition(unitary, backend=None):
    """Decomposes an arbitrary unitary to (A1) from arXiv:quant-ph/0011050."""
    backend = _check_backend(backend)
    unitary = backend.cast(unitary)
    psi, eigvals = calculate_psi(unitary, backend=backend)
    psi_tilde = backend.np.conj(backend.np.sqrt(eigvals)) * backend.np.matmul(
        unitary, psi
    )
    va, vb = calculate_single_qubit_unitaries(psi, backend=backend)
    ua_dagger, ub_dagger = calculate_single_qubit_unitaries(psi_tilde, backend=backend)
    ua, ub = backend.np.transpose(
        backend.np.conj(ua_dagger), (1, 0)
    ), backend.np.transpose(backend.np.conj(ub_dagger), (1, 0))
    return calculate_diagonal(unitary, ua, ub, va, vb, backend=backend)


def to_bell_diagonal(ud, bell_basis=bell_basis, backend=None):
    """Transforms a matrix to the Bell basis and checks if it is diagonal."""
    backend = _check_backend(backend)

    ud = backend.cast(ud)
    bell_basis = backend.cast(bell_basis)

    ud_bell = (
        backend.np.transpose(backend.np.conj(bell_basis), (1, 0)) @ ud @ bell_basis
    )
    ud_diag = backend.np.diag(ud_bell)
    if not np.allclose(backend.np.diag(ud_diag), ud_bell):  # pragma: no cover
        return None
    uprod = backend.np.prod(ud_diag)
    if not np.allclose(uprod, 1):  # pragma: no cover
        return None
    return ud_diag


def calculate_h_vector(ud_diag, backend=None):
    """Finds h parameters corresponding to exp(-iH).

    See Eq. (4)-(5) in arXiv:quant-ph/0307177.
    """
    backend = _check_backend(backend)
    lambdas = -backend.np.angle(ud_diag)
    hx = (lambdas[0] + lambdas[2]) / 2.0
    hy = (lambdas[1] + lambdas[2]) / 2.0
    hz = (lambdas[0] + lambdas[1]) / 2.0
    return hx, hy, hz


def cnot_decomposition(q0, q1, hx, hy, hz, backend=None):
    """Performs decomposition (6) from arXiv:quant-ph/0307177."""
    backend = _check_backend(backend)
    h = backend.cast(H)
    u3 = backend.cast(-1j * matrices.H)
    # use corrected version from PRA paper (not arXiv)
    u2 = -u3 @ gates.RX(0, 2 * hx - np.pi / 2).matrix(backend)
    # add an extra exp(-i pi / 4) global phase to get exact match
    v2 = np.exp(-1j * np.pi / 4) * gates.RZ(0, 2 * hz).matrix(backend)
    v3 = gates.RZ(0, -2 * hy).matrix(backend)
    w = backend.cast((matrices.I - 1j * matrices.X) / np.sqrt(2))
    # change CNOT to CZ using Hadamard gates
    return [
        gates.H(q1),
        gates.CZ(q0, q1),
        gates.Unitary(u2, q0),
        gates.Unitary(h @ v2 @ h, q1),
        gates.CZ(q0, q1),
        gates.Unitary(u3, q0),
        gates.Unitary(h @ v3 @ h, q1),
        gates.CZ(q0, q1),
        gates.Unitary(w, q0),
        gates.Unitary(backend.np.conj(w).T @ h, q1),
    ]


def cnot_decomposition_light(q0, q1, hx, hy, backend=None):
    """Performs decomposition (24) from arXiv:quant-ph/0307177."""
    backend = _check_backend(backend)
    h = backend.cast(H)
    w = backend.cast((matrices.I - 1j * matrices.X) / np.sqrt(2))
    u2 = gates.RX(0, 2 * hx).matrix(backend)
    v2 = gates.RZ(0, -2 * hy).matrix(backend)
    # change CNOT to CZ using Hadamard gates
    return [
        gates.Unitary(backend.np.conj(w).T, q0),
        gates.Unitary(h @ w, q1),
        gates.CZ(q0, q1),
        gates.Unitary(u2, q0),
        gates.Unitary(h @ v2 @ h, q1),
        gates.CZ(q0, q1),
        gates.Unitary(w, q0),
        gates.Unitary(backend.np.conj(w).T @ h, q1),
    ]


def two_qubit_decomposition(q0, q1, unitary, backend=None):
    """Performs two qubit unitary gate decomposition (24) from arXiv:quant-ph/0307177.

    Args:
        q0 (int): index of the first qubit.
        q1 (int): index of the second qubit.
        unitary (ndarray): Unitary :math:`4 \\times 4` to be decomposed.

    Returns:
        (list): gates implementing decomposition (24) from arXiv:quant-ph/0307177
    """
    backend = _check_backend(backend)

    ud_diag = to_bell_diagonal(unitary, backend=backend)
    ud = None
    if ud_diag is None:
        u4, v4, ud, u1, v1 = magic_decomposition(unitary, backend=backend)
        ud_diag = to_bell_diagonal(ud, backend=backend)

    hx, hy, hz = calculate_h_vector(ud_diag, backend=backend)
    hx, hy, hz = float(hx), float(hy), float(hz)
    if np.allclose([hx, hy, hz], [0, 0, 0]):
        u4, v4, ud, u1, v1 = magic_decomposition(unitary, backend=backend)
        gatelist = [gates.Unitary(u4 @ u1, q0), gates.Unitary(v4 @ v1, q1)]
    elif np.allclose(hz, 0):
        gatelist = cnot_decomposition_light(q0, q1, hx, hy, backend=backend)
        if ud is None:
            return gatelist
        g0, g1 = gatelist[:2]
        gatelist[0] = gates.Unitary(backend.cast(g0.parameters[0]) @ u1, q0)
        gatelist[1] = gates.Unitary(backend.cast(g1.parameters[0]) @ v1, q1)

        g0, g1 = gatelist[-2:]
        gatelist[-2] = gates.Unitary(u4 @ g0.parameters[0], q0)
        gatelist[-1] = gates.Unitary(v4 @ g1.parameters[0], q1)

    else:
        cnot_dec = cnot_decomposition(q0, q1, hx, hy, hz, backend=backend)
        if ud is None:
            return cnot_dec

        gatelist = [
            gates.Unitary(u1, q0),
            gates.Unitary(backend.cast(H) @ v1, q1),
        ]
        gatelist.extend(cnot_dec[1:])
        g0, g1 = gatelist[-2:]
        gatelist[-2] = gates.Unitary(u4 @ g0.parameters[0], q0)
        gatelist[-1] = gates.Unitary(v4 @ g1.parameters[0], q1)

    return gatelist



from qibo.config import log, raise_error


def optimize(
    loss,
    initial_parameters,
    args=(),
    method="Powell",
    jac=None,
    hess=None,
    hessp=None,
    bounds=None,
    constraints=(),
    tol=None,
    callback=None,
    options=None,
    compile=False,
    processes=None,
    backend=None,
):
    """Main optimization method. Selects one of the following optimizers:
        - :meth:`qibo.optimizers.cmaes`
        - :meth:`qibo.optimizers.newtonian`
        - :meth:`qibo.optimizers.sgd`

    Args:
        loss (callable): Loss as a function of ``parameters`` and optional extra
            arguments. Make sure the loss function returns a tensor for ``method=sgd``
            and numpy object for all the other methods.
        initial_parameters (np.ndarray): Initial guess for the variational
            parameters that are optimized.
        args (tuple): optional arguments for the loss function.
        method (str): Name of optimizer to use. Can be ``'cma'``, ``'sgd'`` or
            one of the Newtonian methods supported by
            :meth:`qibo.optimizers.newtonian` and ``'parallel_L-BFGS-B'``. ``sgd`` is
            only available for backends based on tensorflow.
        jac (dict): Method for computing the gradient vector for scipy optimizers.
        hess (dict): Method for computing the hessian matrix for scipy optimizers.
        hessp (callable): Hessian of objective function times an arbitrary
            vector for scipy optimizers.
        bounds (sequence or Bounds): Bounds on variables for scipy optimizers.
        constraints (dict): Constraints definition for scipy optimizers.
        tol (float): Tolerance of termination for scipy optimizers.
        callback (callable): Called after each iteration for scipy optimizers.
        options (dict): Dictionary with options. See the specific optimizer
            bellow for a list of the supported options.
        compile (bool): If ``True`` the Tensorflow optimization graph is compiled.
            This is relevant only for the ``'sgd'`` optimizer.
        processes (int): number of processes when using the parallel BFGS method.

    Returns:
        (float, float, custom): Final best loss value; best parameters obtained by the optimizer;         extra: optimizer-specific return object. For scipy methods it
        returns the ``OptimizeResult``, for ``'cma'`` the ``CMAEvolutionStrategy.result``,
        and for ``'sgd'`` the options used during the optimization.


    Example:
        .. testcode::

            import numpy as np
            from qibo import gates, models
            from qibo.optimizers import optimize

            # create custom loss function
            # make sure the return type matches the optimizer requirements.
            def myloss(parameters, circuit):
                circuit.set_parameters(parameters)
                return np.square(np.sum(circuit().state())) # returns numpy array

            # create circuit ansatz for two qubits
            circuit = models.Circuit(2)
            circuit.add(gates.RY(0, theta=0))

            # optimize using random initial variational parameters
            initial_parameters = np.random.uniform(0, 2, 1)
            best, params, extra = optimize(myloss, initial_parameters, args=(circuit))

            # set parameters to circuit
            circuit.set_parameters(params)
    """
    if method == "cma":
        if bounds is not None:  # pragma: no cover
            raise_error(
                RuntimeError,
                "The keyword 'bounds' cannot be used with the cma optimizer. Please use 'options' instead as defined by the cma documentation: ex. options['bounds'] = [0.0, 1.0].",
            )
        return cmaes(loss, initial_parameters, args, callback, options)
    elif method == "sgd":
        from qibo.backends import _check_backend

        backend = _check_backend(backend)

        return sgd(loss, initial_parameters, args, callback, options, compile, backend)
    else:
        from qibo.backends import _check_backend

        backend = _check_backend(backend)

        return newtonian(
            loss,
            initial_parameters,
            args,
            method,
            jac,
            hess,
            hessp,
            bounds,
            constraints,
            tol,
            callback,
            options,
            processes,
            backend,
        )


def cmaes(loss, initial_parameters, args=(), callback=None, options=None):
    """Genetic optimizer based on `pycma <https://github.com/CMA-ES/pycma>`_.

    Args:
        loss (callable): Loss as a function of variational parameters to be
            optimized.
        initial_parameters (np.ndarray): Initial guess for the variational
            parameters.
        args (tuple): optional arguments for the loss function.
        callback (list[callable]): List of callable called after each optimization
            iteration. According to cma-es implementation take ``CMAEvolutionStrategy``
            instance as argument.
            See: https://cma-es.github.io/apidocs-pycma/cma.evolution_strategy.CMAEvolutionStrategy.html.
        options (dict): Dictionary with options accepted by the ``cma``
            optimizer. The user can use ``import cma; cma.CMAOptions()`` to view the
            available options.
    """
    import cma

    es = cma.CMAEvolutionStrategy(initial_parameters, sigma0=1.7, inopts=options)

    if callback is not None:
        while not es.stop():
            solutions = es.ask()
            objective_values = [loss(x, *args) for x in solutions]
            for solution in solutions:
                callback(solution)
            es.tell(solutions, objective_values)
            es.logger.add()
    else:
        es.optimize(loss, args=args)

    return es.result.fbest, es.result.xbest, es.result


def newtonian(
    loss,
    initial_parameters,
    args=(),
    method="Powell",
    jac=None,
    hess=None,
    hessp=None,
    bounds=None,
    constraints=(),
    tol=None,
    callback=None,
    options=None,
    processes=None,
    backend=None,
):
    """Newtonian optimization approaches based on ``scipy.optimize.minimize``.

    For more details check the `scipy documentation <https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.minimize.html>`_.

    .. note::
        When using the method ``parallel_L-BFGS-B`` the ``processes`` option controls the
        number of processes used by the parallel L-BFGS-B algorithm through the ``multiprocessing`` library.
        By default ``processes=None``, in this case the total number of logical cores are used.
        Make sure to select the appropriate number of processes for your computer specification,
        taking in consideration memory and physical cores. In order to obtain optimal results
        you can control the number of threads used by each process with the ``qibo.set_threads`` method.
        For example, for small-medium size circuits you may benefit from single thread per process, thus set
        ``qibo.set_threads(1)`` before running the optimization.

    Args:
        loss (callable): Loss as a function of variational parameters to be
            optimized.
        initial_parameters (np.ndarray): Initial guess for the variational
            parameters.
        args (tuple): optional arguments for the loss function.
        method (str): Name of method supported by ``scipy.optimize.minimize`` and ``'parallel_L-BFGS-B'`` for
            a parallel version of L-BFGS-B algorithm.
        jac (dict): Method for computing the gradient vector for scipy optimizers.
        hess (dict): Method for computing the hessian matrix for scipy optimizers.
        hessp (callable): Hessian of objective function times an arbitrary
            vector for scipy optimizers.
        bounds (sequence or Bounds): Bounds on variables for scipy optimizers.
        constraints (dict): Constraints definition for scipy optimizers.
        tol (float): Tolerance of termination for scipy optimizers.
        callback (callable): Called after each iteration for scipy optimizers.
        options (dict): Dictionary with options accepted by
            ``scipy.optimize.minimize``.
        processes (int): number of processes when using the parallel BFGS method.
    """
    if method == "parallel_L-BFGS-B":  # pragma: no cover
        o = ParallelBFGS(
            loss,
            args=args,
            processes=processes,
            bounds=bounds,
            callback=callback,
            options=options,
        )
        m = o.run(initial_parameters)
    else:
        from scipy.optimize import minimize

        m = minimize(
            loss,
            initial_parameters,
            args=args,
            method=method,
            jac=jac,
            hess=hess,
            hessp=hessp,
            bounds=bounds,
            constraints=constraints,
            tol=tol,
            callback=callback,
            options=options,
        )
    return m.fun, m.x, m


def sgd(
    loss,
    initial_parameters,
    args=(),
    callback=None,
    options=None,
    compile=False,
    backend=None,
):
    """Stochastic Gradient Descent (SGD) optimizer using Tensorflow backpropagation.

    See `tf.keras.Optimizers <https://www.tensorflow.org/api_docs/python/tf/keras/optimizers>`_
    for a list of the available optimizers.

    Args:
        loss (callable): Loss as a function of variational parameters to be
            optimized.
        initial_parameters (np.ndarray): Initial guess for the variational
            parameters.
        args (tuple): optional arguments for the loss function.
        callback (callable): Called after each iteration.
        options (dict): Dictionary with options for the SGD optimizer. Supports
            the following keys:

            - ``'optimizer'`` (str, default: ``'Adagrad'``): Name of optimizer.
            - ``'learning_rate'`` (float, default: ``'1e-3'``): Learning rate.
            - ``'nepochs'`` (int, default: ``1e6``): Number of epochs for optimization.
            - ``'nmessage'`` (int, default: ``1e3``): Every how many epochs to print
              a message of the loss function.
    """

    if not backend.name == "tensorflow":
        raise_error(RuntimeError, "SGD optimizer requires Tensorflow backend.")

    sgd_options = {
        "nepochs": 1000000,
        "nmessage": 1000,
        "optimizer": "Adagrad",
        "learning_rate": 0.001,
    }
    if options is not None:
        sgd_options.update(options)

    # proceed with the training
    vparams = backend.tf.Variable(initial_parameters)
    optimizer = getattr(backend.tf.optimizers, sgd_options["optimizer"])(
        learning_rate=sgd_options["learning_rate"]
    )

    def opt_step():
        with backend.tf.GradientTape() as tape:
            l = loss(vparams, *args)
        grads = tape.gradient(l, [vparams])
        optimizer.apply_gradients(zip(grads, [vparams]))
        return l

    if compile:
        loss = backend.compile(loss)
        opt_step = backend.compile(opt_step)

    for e in range(sgd_options["nepochs"]):
        l = opt_step()
        if callback is not None:
            callback(vparams)
        if e % sgd_options["nmessage"] == 1:
            log.info("ite %d : loss %f", e, l.numpy())

    return loss(vparams, *args).numpy(), vparams.numpy(), sgd_options


class ParallelBFGS:  # pragma: no cover
    """Computes the L-BFGS-B using parallel evaluation using multiprocessing.
    This implementation here is based on https://doi.org/10.32614/RJ-2019-030.

    Args:
        function (function): loss function which returns a numpy object.
        args (tuple): optional arguments for the loss function.
        bounds (list): list of bound values for ``scipy.optimize.minimize`` L-BFGS-B.
        callback (function): function callback ``scipy.optimize.minimize`` L-BFGS-B.
        options (dict): follows ``scipy.optimize.minimize`` syntax for L-BFGS-B.
        processes (int): number of processes when using the paralle BFGS method.
    """

    import numpy as np

    def __init__(
        self,
        function,
        args=(),
        bounds=None,
        callback=None,
        options=None,
        processes=None,
    ):
        self.function = function
        self.args = args
        self.xval = None
        self.function_value = None
        self.jacobian_value = None
        self.precision = self.np.finfo("float64").eps
        self.bounds = bounds
        self.callback = callback
        self.options = options
        self.processes = processes

    def run(self, x0):
        """Executes parallel L-BFGS-B minimization.
        Args:
            x0 (numpy.array): guess for initial solution.

        Returns:
            scipy.minimize result object
        """
        from scipy.optimize import minimize

        out = minimize(
            fun=self.fun,
            x0=x0,
            jac=self.jac,
            method="L-BFGS-B",
            bounds=self.bounds,
            callback=self.callback,
            options=self.options,
        )
        out.hess_inv = out.hess_inv * self.np.identity(len(x0))
        return out

    @staticmethod
    def _eval_approx(eps_at, fun, x, eps):
        if eps_at == 0:
            x_ = x
        else:
            x_ = x.copy()
            if eps_at <= len(x):
                x_[eps_at - 1] += eps
            else:
                x_[eps_at - 1 - len(x)] -= eps
        return fun(x_)

    def evaluate(self, x, eps=1e-8):
        if not (
            self.xval is not None and all(abs(self.xval - x) <= self.precision * 2)
        ):
            eps_at = range(len(x) + 1)
            self.xval = x.copy()

            def operation(epsi):
                return self._eval_approx(
                    epsi, lambda y: self.function(y, *self.args), x, eps
                )

            from joblib import Parallel, delayed

            ret = Parallel(self.processes, prefer="threads")(
                delayed(operation)(epsi) for epsi in eps_at
            )
            self.function_value = ret[0]
            self.jacobian_value = (ret[1 : (len(x) + 1)] - self.function_value) / eps

    def fun(self, x):
        self.evaluate(x)
        return self.function_value

    def jac(self, x):
        self.evaluate(x)
        return self.jacobian_value



"""Module with functions that create random quantum and classical objects."""

import math
import warnings
from functools import cache
from typing import Optional, Union

import numpy as np
from scipy.stats import rv_continuous

from qibo import Circuit, gates, matrices
from qibo.backends import NumpyBackend, _check_backend_and_local_state
from qibo.config import MAX_ITERATIONS, PRECISION_TOL, raise_error
from qibo.quantum_info.basis import comp_basis_to_pauli
from qibo.quantum_info.superoperator_transformations import (
    choi_to_chi,
    choi_to_kraus,
    choi_to_liouville,
    choi_to_pauli,
    choi_to_stinespring,
    vectorization,
)


class _ProbabilityDistributionGaussianLoader(rv_continuous):
    """Probability density function for sampling phases of
    the RBS gates as a function of circuit depth."""

    def _pdf(self, theta: float, depth: int):
        amplitude = 2 * math.gamma(2 ** (depth - 1)) / math.gamma(2 ** (depth - 2)) ** 2

        probability = abs(math.sin(theta) * math.cos(theta)) ** (2 ** (depth - 1) - 1)

        return amplitude * probability / 4


class _probability_distribution_sin(rv_continuous):  # pragma: no cover
    def _pdf(self, theta: float):
        return 0.5 * np.sin(theta)

    def _cdf(self, theta: float):
        return np.sin(theta / 2) ** 2

    def _ppf(self, theta: float):
        return 2 * np.arcsin(np.sqrt(theta))


def uniform_sampling_U3(ngates: int, seed=None, backend=None):
    """Samples parameters for Haar-random :class:`qibo.gates.U3`.

    Args:
        ngates (int): Total number of :math:`U_{3}` gates to be sampled.
        seed (int or :class:`numpy.random.Generator`, optional): Either a generator of random
            numbers or a fixed seed to initialize a generator. If ``None``, initializes
            a generator with a random seed. Default: ``None``.
        backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
            in the execution. If ``None``, it uses :class:`qibo.backends.GlobalBackend`.
            Defaults to ``None``.

    Returns:
        ndarray: array of shape (``ngates``, :math:`3`).
    """
    if not isinstance(ngates, int):
        raise_error(
            TypeError, f"ngates must be type int, but it is type {type(ngates)}."
        )
    elif ngates <= 0:
        raise_error(ValueError, f"ngates must be non-negative, but it is {ngates}.")

    backend, local_state = _check_backend_and_local_state(seed, backend)

    sampler = _probability_distribution_sin(a=0, b=np.pi, seed=local_state)
    phases = local_state.random((ngates, 3))
    phases[:, 0] = sampler.rvs(size=len(phases[:, 0]))
    phases[:, 1] = phases[:, 1] * 2 * np.pi
    phases[:, 2] = phases[:, 2] * 2 * np.pi

    phases = backend.cast(phases, dtype=phases.dtype)

    return phases


def random_gaussian_matrix(
    dims: int,
    rank: Optional[int] = None,
    mean: float = 0,
    stddev: float = 1,
    seed=None,
    backend=None,
):
    """Generates a random Gaussian Matrix.

    Gaussian matrices are matrices where each entry is
    sampled from a Gaussian probability distribution

    .. math::"haar",
        p(x) = \\frac{1}{\\sqrt{2 \\, \\pi} \\, \\sigma} \\,
            \\exp{\\left(-\\frac{(x - \\mu)^{2}}{2\\,\\sigma^{2}}\\right)}

    with mean :math:`\\mu` and standard deviation :math:`\\sigma`.

    Args:
        dims (int): dimension of the matrix.
        rank (int, optional): rank of the matrix. If ``None``, then
            ``rank == dims``. Default: ``None``.
        mean (float, optional): mean of the Gaussian distribution. Defaults to 0.
        stddev (float, optional): standard deviation of the Gaussian distribution.
            Defaults to ``1``.
        seed (int or :class:`numpy.random.Generator`, optional): Either a generator of random
            numbers or a fixed seed to initialize a generator. If ``None``, initializes
            a generator with a random seed. Default: ``None``.
        backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
            in the execution. If ``None``, it uses :class:`qibo.backends.GlobalBackend`.
            Defaults to ``None``.

    Returns:
        ndarray: Random Gaussian matrix with dimensions ``(dims, rank)``.
    """

    if dims <= 0:
        raise_error(ValueError, "dims must be type int and positive.")

    if rank is None:
        rank = dims
    else:
        if rank > dims:
            raise_error(
                ValueError, f"rank ({rank}) cannot be greater than dims ({dims})."
            )
        elif rank <= 0:
            raise_error(ValueError, f"rank ({rank}) must be an int between 1 and dims.")

    if stddev is not None and stddev <= 0.0:
        raise_error(ValueError, "stddev must be a positive float.")

    backend, local_state = _check_backend_and_local_state(seed, backend)

    dims = (dims, rank)

    matrix = 1.0j * local_state.normal(loc=mean, scale=stddev, size=dims)
    matrix += local_state.normal(loc=mean, scale=stddev, size=dims)
    matrix = backend.cast(matrix, dtype=matrix.dtype)

    return matrix


def random_hermitian(
    dims: int,
    semidefinite: bool = False,
    normalize: bool = False,
    seed=None,
    backend=None,
):
    """Generates a random Hermitian matrix :math:`H`, i.e.
    a random matrix such that :math:`H = H^{\\dagger}.`

    Args:
        dims (int): dimension of the matrix.
        semidefinite (bool, optional): if ``True``, returns a Hermitian matrix that
            is also positive semidefinite. Defaults to ``False``.
        normalize (bool, optional): if ``True`` and ``semidefinite=False``, returns
            a Hermitian matrix with eigenvalues in the interval
            :math:`[-1, \\, 1]`. If ``True`` and ``semidefinite=True``,
            interval is :math:`[0, \\, 1]`. Defaults to ``False``.
        seed (int or :class:`numpy.random.Generator`, optional): Either a generator of
            random numbers or a fixed seed to initialize a generator. If ``None``,
            initializes a generator with a random seed. Defaults to ``None``.
        backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
            in the execution. If ``None``, it uses :class:`qibo.backends.GlobalBackend`.
            Defaults to ``None``.

    Returns:
        ndarray: Hermitian matrix :math:`H` with dimensions ``(dims, dims)``.
    """

    if dims <= 0:
        raise_error(ValueError, f"dims ({dims}) must be type int and positive.")

    if not isinstance(semidefinite, bool) or not isinstance(normalize, bool):
        raise_error(TypeError, "semidefinite and normalize must be type bool.")

    backend, local_state = _check_backend_and_local_state(seed, backend)

    matrix = random_gaussian_matrix(dims, dims, seed=local_state, backend=backend)

    if semidefinite:
        matrix = np.dot(np.transpose(np.conj(matrix)), matrix)
    else:
        matrix = (matrix + np.transpose(np.conj(matrix))) / 2

    if normalize:
        matrix = matrix / np.linalg.norm(matrix)

    return matrix


def random_unitary(dims: int, measure: Optional[str] = None, seed=None, backend=None):
    """Returns a random Unitary operator :math:`U`, i.e.
    a random operator such that :math:`U^{-1} = U^{\\dagger}`.

    Args:
        dims (int): dimension of the matrix.
        measure (str, optional): probability measure in which to sample the unitary
            from. If ``None``, functions returns :math:`\\exp{(-i \\, H)}`, where
            :math:`H` is a Hermitian operator. If ``"haar"``, returns an Unitary
            matrix sampled from the Haar measure. Defaults to ``None``.
        seed (int or :class:`numpy.random.Generator`, optional): Either a generator of
            random numbers or a fixed seed to initialize a generator. If ``None``,
            initializes a generator with a random seed. Defaults to ``None``.
        backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
            in the execution. If ``None``, it uses :class:`qibo.backends.GlobalBackend`.
            Defaults to ``None``.

    Returns:
        ndarray: Unitary matrix :math:`U` with dimensions ``(dims, dims)``.
    """

    if dims <= 0:
        raise_error(ValueError, "dims must be type int and positive.")

    if measure is not None:
        if not isinstance(measure, str):
            raise_error(
                TypeError, f"measure must be type str but it is type {type(measure)}."
            )
        if measure != "haar":
            raise_error(ValueError, f"measure {measure} not implemented.")

    backend, local_state = _check_backend_and_local_state(seed, backend)

    if measure == "haar":
        unitary = random_gaussian_matrix(dims, dims, seed=local_state, backend=backend)
        Q, R = np.linalg.qr(unitary)
        D = np.diag(R)
        D = D / np.abs(D)
        R = np.diag(D)
        unitary = np.dot(Q, R)
    elif measure is None:
        from scipy.linalg import expm

        H = random_hermitian(dims, seed=seed, backend=NumpyBackend())
        unitary = expm(-1.0j * H / 2)

    unitary = backend.cast(unitary, dtype=unitary.dtype)

    return unitary


def random_quantum_channel(
    dims: int,
    representation: str = "liouville",
    measure: Optional[str] = None,
    rank: Optional[int] = None,
    order: str = "row",
    normalize: bool = False,
    precision_tol: Optional[float] = None,
    validate_cp: bool = True,
    nqubits: Optional[int] = None,
    initial_state_env=None,
    seed=None,
    backend=None,
):
    """Creates a random superoperator from an unitary operator in one of the
    supported superoperator representations.

    Args:
        dims (int): dimension of the :math:`n`-qubit operator, i.e. :math:`\\text{dims}=2^{n}`.
        representation (str, optional): If ``"chi"``, returns a random channel in the
            Chi representation. If ``"choi"``, returns channel in Choi representation.
            If ``"kraus"``, returns Kraus representation of channel. If ``"liouville"``,
            returns Liouville representation. If ``"pauli"``, returns Pauli-Liouville
            representation. If "pauli-<pauli_order>" or "chi-<pauli_order>", (e.g. "pauli-IZXY"),
            returns it in the Pauli basis with the corresponding order of single-qubit Pauli elements
            (see :func:`qibo.quantum_info.pauli_basis`). If ``"stinespring"``,
            returns random channel in the Stinespring representation. Defaults to ``"liouville"``.
        measure (str, optional): probability measure in which to sample the unitary
            from. If ``None``, functions returns :math:`\\exp{(-i \\, H)}`, where
            :math:`H` is a Hermitian operator. If ``"haar"``, returns an Unitary
            matrix sampled from the Haar measure. If ``"bcsz"``, it samples an unitary
            from the BCSZ distribution with Kraus ``rank``. Defaults to ``None``.
        rank (int, optional): used when ``measure=="bcsz"``. Rank of the matrix.
            If ``None``, then ``rank==dims``. Defaults to ``None``.
        order (str, optional): If ``"row"``, vectorization is performed row-wise.
            If ``"column"``, vectorization is performed column-wise. If ``"system"``,
            a block-vectorization is performed. Defaults to ``"row"``.
        normalize (bool, optional): used when ``representation="chi"`` or
            ``representation="pauli"``. If ``True`` assumes the normalized Pauli basis.
            If ``False``, it assumes unnormalized Pauli basis. Defaults to ``False``.
        precision_tol (float, optional): if ``representation="kraus"``, it is the
            precision tolerance for eigenvalues found in the spectral decomposition
            problem. Any eigenvalue :math:`\\lambda <` ``precision_tol`` is set
            to 0 (zero). If ``None``, ``precision_tol`` defaults to
            ``qibo.config.PRECISION_TOL=1e-8``. Defaults to ``None``.
        validate_cp (bool, optional): used when ``representation="stinespring"``.
            If ``True``, checks if the Choi representation of superoperator
            used as intermediate step is a completely positive map.
            If ``False``, it assumes that it is completely positive (and Hermitian).
            Defaults to ``True``.
        nqubits (int, optional): used when ``representation="stinespring"``.
            Total number of qubits in the system that is interacting with
            the environment. Must be equal or greater than the number of
            qubits that Kraus representation of the system superoperator acts on.
            If ``None``, defaults to the number of qubits in the Kraus operators.
            Defauts to ``None``.
        initial_state_env (ndarray, optional): used when ``representation="stinespring"``.
            Statevector representing the initial state of the enviroment.
            If ``None``, it assumes the environment in its ground state.
            Defaults to ``None``.
        seed (int or :class:`numpy.random.Generator`, optional): Either a generator of
            random numbers or a fixed seed to initialize a generator. If ``None``,
            initializes a generator with a random seed. Defaults to ``None``.
        backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
            in the execution. If ``None``, it uses :class:`qibo.backends.GlobalBackend`.
            Defaults to ``None``.

    Returns:
        ndarray: Superoperator representation of a random unitary gate.
    """
    if not isinstance(representation, str):
        raise_error(
            TypeError,
            f"representation must be type str, but it is type {type(representation)}",
        )

    if representation not in [
        "chi",
        "choi",
        "kraus",
        "liouville",
        "pauli",
        "stinespring",
    ]:
        if (
            ("chi-" not in representation and "pauli-" not in representation)
            or len(representation.split("-")) != 2
            or set(representation.split("-")[1]) != {"I", "X", "Y", "Z"}
        ):
            raise_error(ValueError, f"representation {representation} not implemented.")

    if measure == "bcsz" and order not in ["row", "column"]:
        raise_error(
            NotImplementedError, f"order {order} not implemented for measure {measure}."
        )

    backend, local_state = _check_backend_and_local_state(seed, backend)

    if measure == "bcsz":
        super_op = _super_op_from_bcsz_measure(
            dims=dims, rank=rank, order=order, seed=local_state, backend=backend
        )
    else:
        super_op = random_unitary(dims, measure, local_state, backend)
        super_op = vectorization(super_op, order=order, backend=backend)
        super_op = np.outer(super_op, np.conj(super_op))

    if "chi" in representation:
        pauli_order = "IXYZ"
        if "-" in representation:
            pauli_order = representation.split("-")[1]
        super_op = choi_to_chi(
            super_op,
            normalize=normalize,
            order=order,
            pauli_order=pauli_order,
            backend=backend,
        )
    elif representation == "kraus":
        super_op = choi_to_kraus(
            super_op,
            precision_tol=precision_tol,
            order=order,
            validate_cp=False,
            backend=backend,
        )
    elif representation == "liouville":
        super_op = choi_to_liouville(super_op, order=order, backend=backend)
    elif "pauli" in representation:
        pauli_order = "IXYZ"
        if "-" in representation:
            pauli_order = representation.split("-")[1]
        super_op = choi_to_pauli(
            super_op,
            normalize=normalize,
            order=order,
            pauli_order=pauli_order,
            backend=backend,
        )
    elif representation == "stinespring":
        super_op = choi_to_stinespring(
            super_op,
            precision_tol=precision_tol,
            order=order,
            validate_cp=validate_cp,
            nqubits=nqubits,
            initial_state_env=initial_state_env,
            backend=backend,
        )

    return super_op


def random_statevector(dims: int, seed=None, backend=None):
    """Creates a random statevector :math:`\\ket{\\psi}`.

    .. math::
        \\ket{\\psi} = \\sum_{k = 0}^{d - 1} \\, \\sqrt{p_{k}} \\,
            e^{i \\phi_{k}} \\, \\ket{k} \\, ,

    where :math:`d` is ``dims``, and :math:`p_{k}` and :math:`\\phi_{k}` are, respectively,
    the probability and phase corresponding to the computational basis state :math:`\\ket{k}`.

    Args:
        dims (int): dimension of the matrix.
        seed (int or :class:`numpy.random.Generator`, optional): Either a generator of
            random numbers or a fixed seed to initialize a generator. If ``None``,
            initializes a generator with a random seed. Defaults to ``None``.
        backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
            in the execution. If ``None``, it uses :class:`qibo.backends.GlobalBackend`.
            Defaults to ``None``.

    Returns:
        ndarray: Random statevector :math:`\\ket{\\psi}`.
    """

    if dims <= 0:
        raise_error(ValueError, "dim must be of type int and >= 1")

    if (
        seed is not None
        and not isinstance(seed, int)
        and not isinstance(seed, np.random.Generator)
    ):
        raise_error(
            TypeError, "seed must be either type int or numpy.random.Generator."
        )

    backend, local_state = _check_backend_and_local_state(seed, backend)

    state = local_state.standard_normal(dims).astype(complex)
    state += 1.0j * local_state.standard_normal(dims)
    state /= np.linalg.norm(state)
    state = backend.cast(state, dtype=state.dtype)

    return state


def random_density_matrix(
    dims: int,
    rank: Optional[int] = None,
    pure: bool = False,
    metric: str = "hilbert-schmidt",
    basis: Optional[str] = None,
    normalize: bool = False,
    order: str = "row",
    seed=None,
    backend=None,
):
    """Creates a random density matrix :math:`\\rho`. If ``pure=True``,

    .. math::
        \\rho = \\ketbra{\\psi}{\\psi} \\, ,

    where :math:`\\ket{\\psi}` is a :func:`qibo.quantum_info.random_statevector`.
    If ``pure=False``, then

    .. math::
        \\rho = \\sum_{k} \\, p_{k} \\, \\ketbra{\\psi_{k}}{\\psi_{k}} \\, .

    is a mixed state.

    Args:
        dims (int): dimension of the matrix.
        rank (int, optional): rank of the matrix. If ``None``, then ``rank == dims``.
            Defaults to ``None``.
        pure (bool, optional): if ``True``, returns a pure state. Defaults to ``False``.
        metric (str, optional): metric to sample the density matrix from. Options:
            ``"hilbert-schmidt"``, ``"ginibre"``, and ``"bures"``.
            Note that, by definition, ``rank`` defaults to ``None``
            when ``metric=="hilbert-schmidt"``. Defaults to ``"hilbert-schmidt"``.
        basis (str, optional): if ``None``, returns random density matrix in the
            computational basis. If ``"pauli-<pauli_order>"``, (e.g. ``"pauli-IZXY"``),
            returns it in the Pauli basis with the corresponding order of single-qubit
            Pauli elements (see :func:`qibo.quantum_info.pauli_basis`).
            Defaults to ``None``.
        normalize(bool, optional): used when ``basis="pauli-<pauli-order>"``. If ``True``
            returns random density matrix in the normalized Pauli basis. If ``False``,
            returns state in the unnormalized Pauli basis. Defaults to ``False``.
        order (str, optional): used when ``basis="pauli-<pauli-order>"``. If ``"row"``,
            vectorization of Pauli basis is performed row-wise. If ``"column"``,
            vectorization is performed column-wise. If ``"system"``, system-wise
            vectorization is performed. Default is ``"row"``.
        seed (int or :class:`numpy.random.Generator`, optional): Either a generator of
            random numbers or a fixed seed to initialize a generator. If ``None``,
            initializes a generator with a random seed. Defaults to ``None``.
        backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
            in the execution. If ``None``, it uses :class:`qibo.backends.GlobalBackend`.
            Defaults to ``None``.

    Returns:
        ndarray: Random density matrix :math:`\\rho`.
    """

    if dims <= 0:
        raise_error(ValueError, "dims must be type int and positive.")

    if rank is not None and rank > dims:
        raise_error(ValueError, f"rank ({rank}) cannot be greater than dims ({dims}).")

    if rank is not None and rank <= 0:
        raise_error(ValueError, f"rank ({rank}) must be an int between 1 and dims.")

    if rank is not None and not isinstance(rank, int):
        raise_error(TypeError, f"rank must be type int, but it is type {type(rank)}.")

    if not isinstance(pure, bool):
        raise_error(TypeError, f"pure must be type bool, but it is type {type(pure)}.")

    if not isinstance(metric, str):
        raise_error(
            TypeError, f"metric must be type str, but it is type {type(metric)}."
        )
    if metric not in ["hilbert-schmidt", "ginibre", "bures"]:
        raise_error(ValueError, f"metric {metric} not implemented.")

    if basis is not None and not isinstance(basis, str):
        raise_error(TypeError, f"basis must be type str, but it is type {type(basis)}.")
    elif basis is not None and basis not in ["pauli"]:
        if (
            "pauli-" not in basis
            or len(basis.split("-")) != 2
            or set(basis.split("-")[1]) != {"I", "X", "Y", "Z"}
        ):
            raise_error(ValueError, f"basis {basis} nor recognized.")

    if not isinstance(normalize, bool):
        raise_error(
            TypeError, f"normalize must be type bool, but it is type {type(normalize)}."
        )
    elif normalize is True and basis is None:
        raise_error(ValueError, "normalize cannot be True when basis=None.")

    backend, local_state = _check_backend_and_local_state(seed, backend)

    if metric == "hilbert-schmidt":
        rank = None

    if pure:
        state = random_statevector(dims, seed=local_state, backend=backend)
        state = np.outer(state, np.transpose(np.conj(state)))
    else:
        if metric in ["hilbert-schmidt", "ginibre"]:
            state = random_gaussian_matrix(
                dims, rank, mean=0, stddev=1, seed=local_state, backend=backend
            )
            state = np.dot(state, np.transpose(np.conj(state)))
            state = state / np.trace(state)
        else:
            nqubits = int(np.log2(dims))
            state = backend.identity_density_matrix(nqubits, normalize=False)
            state += random_unitary(dims, seed=local_state, backend=backend)
            state = np.dot(
                state,
                random_gaussian_matrix(dims, rank, seed=local_state, backend=backend),
            )
            state = np.dot(state, np.transpose(np.conj(state)))
            state /= np.trace(state)

    state = backend.cast(state, dtype=state.dtype)

    if basis is not None:
        pauli_order = basis.split("-")[1]
        unitary = comp_basis_to_pauli(
            int(np.log2(dims)),
            normalize=normalize,
            order=order,
            pauli_order=pauli_order,
            backend=backend,
        )
        state = unitary @ vectorization(state, order=order, backend=backend)

    return state


def random_clifford(
    nqubits: int,
    return_circuit: bool = True,
    density_matrix: bool = False,
    seed=None,
    backend=None,
):
    """Generates a random :math:`n`-qubit Clifford operator, where :math:`n` is ``nqubits``.
    For the mathematical details, see Reference [1].

    Args:
        nqubits (int): number of qubits.
        return_circuit (bool, optional): if ``True``, returns a :class:`qibo.models.Circuit`
            object. If ``False``, returns an ``ndarray`` object. Defaults to ``True``.
        density_matrix (bool, optional): used when ``return_circuit=True``. If `True`,
            the circuit would evolve density matrices. Defaults to ``False``.
        seed (int or :class:`numpy.random.Generator`, optional): Either a generator of
            random numbers or a fixed seed to initialize a generator. If ``None``,
            initializes a generator with a random seed. Defaults to ``None``.
        backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
            in the execution. If ``None``, it uses :class:`qibo.backends.GlobalBackend`.
            Defaults to ``None``.

    Returns:
        (ndarray or :class:`qibo.models.Circuit`): Random Clifford operator.

    Reference:
        1. S. Bravyi and D. Maslov, *Hadamard-free circuits expose the
           structure of the Clifford group*.
           `arXiv:2003.09412 [quant-ph] <https://arxiv.org/abs/2003.09412>`_.
    """

    if isinstance(nqubits, int) is False:
        raise_error(
            TypeError,
            f"nqubits must be type int, but it is type {type(nqubits)}.",
        )

    if nqubits <= 0:
        raise_error(ValueError, "nqubits must be a positive integer.")

    if not isinstance(return_circuit, bool):
        raise_error(
            TypeError,
            f"return_circuit must be type bool, but it is type {type(return_circuit)}.",
        )

    backend, local_state = _check_backend_and_local_state(seed, backend)

    hadamards, permutations = _sample_from_quantum_mallows_distribution(
        nqubits, local_state=local_state
    )

    delta_matrix = np.eye(nqubits, dtype=int)
    delta_matrix_prime = np.copy(delta_matrix)

    gamma_matrix_prime = local_state.integers(0, 2, size=nqubits)
    gamma_matrix_prime = np.diag(gamma_matrix_prime)

    gamma_matrix = local_state.integers(0, 2, size=nqubits)
    gamma_matrix = hadamards * gamma_matrix
    gamma_matrix = np.diag(gamma_matrix)

    # filling off-diagonal elements of gammas and deltas matrices
    for j in range(nqubits):
        for k in range(j + 1, nqubits):
            b = local_state.integers(0, 2)
            gamma_matrix_prime[k, j] = b
            gamma_matrix_prime[j, k] = b

            b = local_state.integers(0, 2)
            delta_matrix_prime[k, j] = b

            if hadamards[k] == 1 and hadamards[j] == 1:  # pragma: no cover
                b = local_state.integers(0, 2)
                gamma_matrix[k, j] = b
                gamma_matrix[j, k] = b
                if permutations[k] > permutations[j]:
                    b = local_state.integers(0, 2)
                    delta_matrix[k, j] = b

            if hadamards[k] == 0 and hadamards[j] == 1:
                b = local_state.integers(0, 2)
                delta_matrix[k, j] = b
                if permutations[k] > permutations[j]:
                    b = local_state.integers(0, 2)
                    gamma_matrix[k, j] = b
                    gamma_matrix[j, k] = b

            if (
                hadamards[k] == 1
                and hadamards[j] == 0
                and permutations[k] < permutations[j]
            ):  # pragma: no cover
                b = local_state.integers(0, 2)
                gamma_matrix[k, j] = b
                gamma_matrix[j, k] = b

            if (
                hadamards[k] == 0
                and hadamards[j] == 0
                and permutations[k] < permutations[j]
            ):  # pragma: no cover
                b = local_state.integers(0, 2)
                delta_matrix[k, j] = b

    # get first element of the Borel group
    clifford_circuit = _operator_from_hadamard_free_group(
        gamma_matrix, delta_matrix, density_matrix
    )

    # Apply permutated Hadamard layer
    for qubit, had in enumerate(hadamards):
        if had == 1:
            clifford_circuit.add(gates.H(int(permutations[qubit])))

    # get second element of the Borel group
    clifford_circuit += _operator_from_hadamard_free_group(
        gamma_matrix_prime,
        delta_matrix_prime,
        density_matrix,
        random_pauli(
            nqubits,
            depth=1,
            return_circuit=True,
            density_matrix=density_matrix,
            seed=local_state,
            backend=backend,
        ),
    )

    if return_circuit is False:
        clifford_circuit = clifford_circuit.unitary(backend=backend)

    return clifford_circuit


def random_pauli(
    qubits,
    depth: int,
    max_qubits: Optional[int] = None,
    subset: Optional[list] = None,
    return_circuit: bool = True,
    density_matrix: bool = False,
    seed=None,
    backend=None,
):
    """Creates random Pauli operator(s).

    Pauli operators are sampled from the single-qubit Pauli set
    :math:`\\{I, \\, X, \\, Y, \\, Z\\}`.

    Args:
        qubits (int or list or ndarray): if ``int`` and ``max_qubits=None``, the
            number of qubits. If ``int`` and ``max_qubits != None``, qubit index
            in which the Pauli sequence will act. If ``list`` or ``ndarray``,
            indexes of the qubits for the Pauli sequence to act.
        depth (int): length of the sequence of Pauli gates.
        max_qubits (int, optional): total number of qubits in the circuit.
            If ``None``, ``max_qubits = max(qubits)``. Defaults to ``None``.
        subset (list, optional): list containing a subset of the 4 single-qubit
            Pauli operators. If ``None``, defaults to the complete set.
            Defaults to ``None``.
        return_circuit (bool, optional): if ``True``, returns a :class:`qibo.models.Circuit`
            object. If ``False``, returns an ``ndarray`` with shape (qubits, depth, 2, 2)
            that contains all Pauli matrices that were sampled. Defaults to ``True``.
        density_matrix (bool, optional): used when ``return_circuit=True``. If `True`,
            the circuit would evolve density matrices. Defaults to ``False``.
        seed (int or :class:`numpy.random.Generator`, optional): Either a generator of
            random numbers or a fixed seed to initialize a generator. If ``None``,
            initializes a generator with a random seed. Defaults to ``None``.
        backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
            in the execution. If ``None``, it uses :class:`qibo.backends.GlobalBackend`.
            Defaults to ``None``.

    Returns:
        (ndarray or :class:`qibo.models.Circuit`): all sampled Pauli operators.

    """

    if (
        not isinstance(qubits, int)
        and not isinstance(qubits, list)
        and not isinstance(qubits, np.ndarray)
    ):
        raise_error(
            TypeError,
            f"qubits must be either type int, list or ndarray, but it is type {type(qubits)}.",
        )

    if isinstance(qubits, int) and qubits < 0:
        raise_error(ValueError, "qubits must be a non-negative integer.")

    if isinstance(qubits, int) is False and any(q < 0 for q in qubits):
        raise_error(ValueError, "qubit indexes must be non-negative integers.")

    if isinstance(depth, int) and depth <= 0:
        raise_error(ValueError, "depth must be a positive integer.")

    if isinstance(max_qubits, int) and max_qubits <= 0:
        raise_error(ValueError, "max_qubits must be a positive integer.")

    if max_qubits is not None:
        if isinstance(qubits, int) and qubits >= max_qubits:
            raise_error(
                ValueError,
                f"qubit index ({qubits}) must be < max_qubits ({max_qubits}).",
            )
        elif not isinstance(qubits, int) and any(q >= max_qubits for q in qubits):
            raise_error(ValueError, "all qubit indexes must be < max_qubits.")

    if not isinstance(return_circuit, bool):
        raise_error(
            TypeError,
            f"return_circuit must be type bool, but it is type {type(return_circuit)}.",
        )

    if subset is not None and not isinstance(subset, list):
        raise_error(
            TypeError, f"subset must be type list, but it is type {type(subset)}."
        )

    if subset is not None and any(isinstance(item, str) is False for item in subset):
        raise_error(
            TypeError,
            "subset argument must be a subset of strings in the set ['I', 'X', 'Y', 'Z'].",
        )

    backend, local_state = _check_backend_and_local_state(seed, backend)

    complete_set = (
        {"I": gates.I, "X": gates.X, "Y": gates.Y, "Z": gates.Z}
        if return_circuit
        else {"I": matrices.I, "X": matrices.X, "Y": matrices.Y, "Z": matrices.Z}
    )

    if subset is None:
        subset = complete_set
    else:
        subset = {key: complete_set[key] for key in subset}

    keys = list(subset.keys())

    if max_qubits is None:
        if isinstance(qubits, int):
            max_qubits = qubits
            qubits = range(qubits)
        else:
            max_qubits = int(max(qubits)) + 1
    else:
        if isinstance(qubits, int):
            qubits = [qubits]

    indexes = local_state.integers(0, len(subset), size=(len(qubits), depth))
    indexes = [[keys[item] for item in row] for row in indexes]

    if return_circuit:
        gate_grid = Circuit(max_qubits, density_matrix=density_matrix)
        for qubit, row in zip(qubits, indexes):
            for column_item in row:
                if subset[column_item] != gates.I:
                    gate_grid.add(subset[column_item](qubit))
    else:
        gate_grid = backend.cast(
            [[subset[column_item] for column_item in row] for row in indexes]
        )

    return gate_grid


def random_pauli_hamiltonian(
    nqubits: int,
    max_eigenvalue: Optional[Union[int, float]] = None,
    normalize: bool = False,
    pauli_order: str = "IXYZ",
    seed=None,
    backend=None,
):
    """Generates a random Hamiltonian in the Pauli basis.

    Args:
        nqubits (int): number of qubits.
        max_eigenvalue (int or float, optional): fixes the value of the
            largest eigenvalue. Defaults to ``None``.
        normalize (bool, optional): If ``True``, fixes the gap of the
            Hamiltonian as ``1.0``. Moreover, if ``True``, then ``max_eigenvalue``
            must be ``> 1.0``. Defaults to ``False``.
        pauli_order (str, optional): corresponds to the order of 4 single-qubit
            Pauli elements in the basis. Defaults to "IXYZ".
        seed (int or :class:`numpy.random.Generator`, optional): Either a generator of
            random numbers or a fixed seed to initialize a generator. If ``None``,
            initializes a generator with a random seed. Defaults to ``None``.
        backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
            in the execution. If ``None``, it uses :class:`qibo.backends.GlobalBackend`.
            Defaults to ``None``.

    Returns:
        tuple(ndarray, ndarray): Hamiltonian in the Pauli basis and its corresponding eigenvalues.
    """
    if isinstance(nqubits, int) is False:
        raise_error(
            TypeError, f"nqubits must be type int, but it is type {type(nqubits)}."
        )
    elif nqubits <= 0:
        raise_error(ValueError, "nqubits must be a positive int.")

    if isinstance(max_eigenvalue, (int, float)) is False and normalize is True:
        raise_error(
            TypeError,
            f"when normalize=True, max_eigenvalue must be type float, "
            + f"but it is {type(max_eigenvalue)}.",
        )
    elif (
        isinstance(max_eigenvalue, (int, float)) is False and max_eigenvalue is not None
    ):
        raise_error(
            TypeError,
            f"max_eigenvalue must be type float, but it is {type(max_eigenvalue)}.",
        )

    if isinstance(normalize, bool) is False:
        raise_error(
            TypeError,
            f"normalize must be type bool, but it is type {type(normalize)}.",
        )
    elif normalize is True and max_eigenvalue <= 1.0:
        raise_error(
            ValueError,
            "when normalize=True, gap is = 1, thus max_eigenvalue must be > 1.",
        )

    backend, local_state = _check_backend_and_local_state(seed, backend)

    d = 2**nqubits

    hamiltonian = random_hermitian(d, normalize=True, seed=local_state, backend=backend)

    eigenvalues, eigenvectors = np.linalg.eigh(hamiltonian)

    if normalize is True:
        eigenvalues -= eigenvalues[0]

        eigenvalues /= eigenvalues[1]

        shift = 2
        eigenvectors[:, shift:] = (
            eigenvectors[:, shift:] * max_eigenvalue / eigenvalues[-1]
        )
        eigenvalues[shift:] = eigenvalues[shift:] * max_eigenvalue / eigenvalues[-1]

        hamiltonian = np.zeros((d, d), dtype=complex)
        hamiltonian = backend.cast(hamiltonian, dtype=hamiltonian.dtype)
        # excluding the first eigenvector because first eigenvalue is zero
        for eigenvalue, eigenvector in zip(
            eigenvalues[1:], np.transpose(eigenvectors)[1:]
        ):
            hamiltonian += eigenvalue * np.outer(eigenvector, np.conj(eigenvector))

    U = comp_basis_to_pauli(
        nqubits, normalize=True, pauli_order=pauli_order, backend=backend
    )

    hamiltonian = np.real(U @ vectorization(hamiltonian, backend=backend))

    return hamiltonian, eigenvalues


def random_stochastic_matrix(
    dims: int,
    bistochastic: bool = False,
    diagonally_dominant: bool = False,
    precision_tol: Optional[float] = None,
    max_iterations: Optional[int] = None,
    seed=None,
    backend=None,
):
    """Creates a random stochastic matrix.

    Args:
        dims (int): dimension of the matrix.
        bistochastic (bool, optional): if ``True``, matrix is row- and column-stochastic.
            If ``False``, matrix is row-stochastic. Defaults to ``False``.
        diagonally_dominant (bool, optional): if ``True``, matrix is strictly diagonally
            dominant. Defaults to ``False``.
        precision_tol (float, optional): tolerance level for how much each probability
            distribution can deviate from summing up to ``1.0``. If ``None``,
            it defaults to ``qibo.config.PRECISION_TOL``. Defaults to ``None``.
        max_iterations (int, optional): when ``bistochastic=True``, maximum number
            of iterations used to normalize all rows and columns simultaneously.
            If ``None``, defaults to ``qibo.config.MAX_ITERATIONS``.
            Defaults to ``None``.
        seed (int or :class:`numpy.random.Generator`, optional): Either a generator of
            random numbers or a fixed seed to initialize a generator. If ``None``,
            initializes a statevectorgenerator with a random seed. Defaults to ``None``.
        backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
            in the execution. If ``None``, it uses :class:`qibo.backends.GlobalBackend`.
            Defaults to ``None``.

    Returns:
        ndarray: a random stochastic matrix.

    """
    if dims <= 0:
        raise_error(ValueError, "dims must be type int and positive.")

    if not isinstance(bistochastic, bool):
        raise_error(
            TypeError,
            f"bistochastic must be type bool, but it is type {type(bistochastic)}.",
        )

    if not isinstance(diagonally_dominant, bool):
        raise_error(
            TypeError,
            f"diagonally_dominant must be type bool, but it is type {type(diagonally_dominant)}.",
        )

    if precision_tol is not None:
        if not isinstance(precision_tol, float):
            raise_error(
                TypeError,
                f"precision_tol must be type float, but it is type {type(precision_tol)}.",
            )
        if precision_tol < 0.0:
            raise_error(ValueError, "precision_tol must be non-negative.")

    if max_iterations is not None:
        if not isinstance(max_iterations, int):
            raise_error(
                TypeError,
                f"max_iterations must be type int, but it is type {type(precision_tol)}.",
            )
        if max_iterations <= 0.0:
            raise_error(ValueError, "max_iterations must be a positive int.")

    backend, local_state = _check_backend_and_local_state(seed, backend)

    if precision_tol is None:
        precision_tol = PRECISION_TOL
    if max_iterations is None:
        max_iterations = MAX_ITERATIONS

    matrix = local_state.random(size=(dims, dims))
    if diagonally_dominant:
        matrix /= dims**2
        for k, row in enumerate(matrix):
            row = np.delete(row, obj=k)
            matrix[k, k] = 1 - np.sum(row)
    row_sum = np.sum(matrix, axis=1)

    row_sum = matrix.sum(axis=1)

    if bistochastic:
        column_sum = matrix.sum(axis=0)
        count = 0
        while count <= max_iterations - 1 and (
            (
                np.any(row_sum >= 1 + precision_tol)
                or np.any(row_sum <= 1 - precision_tol)
            )
            or (
                np.any(column_sum >= 1 + precision_tol)
                or np.any(column_sum <= 1 - precision_tol)
            )
        ):
            matrix = matrix / matrix.sum(axis=0)
            matrix = matrix / matrix.sum(axis=1)[:, np.newaxis]
            row_sum = matrix.sum(axis=1)
            column_sum = matrix.sum(axis=0)
            count += 1
        if count == max_iterations:
            warnings.warn("Reached max iterations.", RuntimeWarning)
    else:
        matrix = matrix / np.outer(row_sum, [1] * dims)

    matrix = backend.cast(matrix, dtype=matrix.dtype)

    return matrix


def _sample_from_quantum_mallows_distribution(nqubits: int, local_state):
    """Using the quantum Mallows distribution, samples a binary array
    representing a layer of Hadamard gates as well as an array with permutated
    qubit indexes. For more details, see Reference [1].

    Args:
        nqubits (int): number of qubits.
        local_state (:class:`numpy.random.Generator`): a generator of
            random numbers

    Returns:
        (``ndarray``, ``ndarray`): tuple of binary ``ndarray`` and ``ndarray`` of indexes.

    Reference:
        1. S. Bravyi and D. Maslov, *Hadamard-free circuits expose the
            structure of the Clifford group*.
            `arXiv:2003.09412 [quant-ph] <https://arxiv.org/abs/2003.09412>`_.

    """
    mute_index = list(range(nqubits))

    exponents = np.arange(nqubits, 0, -1, dtype=np.int64)
    powers = 4**exponents
    powers[powers == 0] = np.iinfo(np.int64).max

    r = local_state.uniform(0, 1, size=nqubits)

    indexes = -1 * (np.ceil(np.log2(r + (1 - r) / powers)))

    hadamards = 1 * (indexes < exponents)

    permutations = np.zeros(nqubits, dtype=int)
    for l, (index, m) in enumerate(zip(indexes, exponents)):
        k = index if index < m else 2 * m - index - 1
        k = int(k)
        permutations[l] = mute_index[k]
        del mute_index[k]

    return hadamards, permutations


@cache
def _create_S(q):
    return gates.S(int(q))


@cache
def _create_CZ(cq, tq):
    return gates.CZ(int(cq), int(tq))


@cache
def _create_CNOT(cq, tq):
    return gates.CNOT(int(cq), int(tq))


def _operator_from_hadamard_free_group(
    gamma_matrix, delta_matrix, density_matrix: bool = False, pauli_operator=None
):
    """Calculates an element :math:`F` of the Hadamard-free group :math:`\\mathcal{F}_{n}`,
    where :math:`n` is the number of qubits ``nqubits``. For more details,
    see Reference [1].

    Args:
        gamma_matrix (ndarray): :math:`\\, n \\times n \\,` binary matrix.
        delta_matrix (ndarray): :math:`\\, n \\times n \\,` binary matrix.
        density_matrix (bool, optional): used when ``return_circuit=True``. If `True`,
            the circuit would evolve density matrices. Defaults to ``False``.
        pauli_operator (:class:`qibo.models.Circuit`, optional): a :math:`n`-qubit
            Pauli operator. If ``None``, it is assumed to be the Identity.
            Defaults to ``None``.

    Returns:
        :class:`qibo.models.Circuit`: element of the Hadamard-free group.

    Reference:
        1. S. Bravyi and D. Maslov, *Hadamard-free circuits expose the
           structure of the Clifford group*.
           `arXiv:2003.09412 [quant-ph] <https://arxiv.org/abs/2003.09412>`_.
    """
    if gamma_matrix.shape != delta_matrix.shape:  # pragma: no cover
        raise_error(
            ValueError,
            "gamma_matrix and delta_matrix must have shape (nqubits, nqubits), "
            + f"but {gamma_matrix.shape} != {delta_matrix.shape}",
        )

    nqubits = len(gamma_matrix)
    circuit = Circuit(nqubits, density_matrix=density_matrix)

    if pauli_operator is not None:
        circuit += pauli_operator

    idx = np.tril_indices(nqubits, k=-1)
    gamma_ones = gamma_matrix[idx].nonzero()[0]
    delta_ones = delta_matrix[idx].nonzero()[0]

    S_gates = [_create_S(q) for q in np.diag(gamma_matrix).nonzero()[0]]
    CZ_gates = [
        _create_CZ(cq, tq) for cq, tq in zip(idx[1][gamma_ones], idx[0][gamma_ones])
    ]
    CNOT_gates = [
        _create_CNOT(cq, tq) for cq, tq in zip(idx[1][delta_ones], idx[0][delta_ones])
    ]

    circuit.add(S_gates + CZ_gates + CNOT_gates)

    return circuit


def _super_op_from_bcsz_measure(dims: int, rank: int, order: str, seed, backend):
    """Helper function for :func:qibo.quantum_info.random_ensembles.random_quantum_channel.
    Generates a channel from the BCSZ measure.

    Args:
        dims (int): dimension of the :math:`n`-qubit operator, i.e. :math:`\\text{dims}=2^{n}`.
        rank (int, optional): used when ``measure=="bcsz"``. Rank of the matrix.
            If ``None``, then ``rank==dims``. Defaults to ``None``.
        order (str, optional): If ``"row"``, vectorization is performed row-wise.
            If ``"column"``, vectorization is performed column-wise. Defaults to ``"row"``.
        seed (int or :class:`numpy.random.Generator`, optional): Either a generator of
            random numbers or a fixed seed to initialize a generator. If ``None``,
            initializes a generator with a random seed. Defaults to ``None``.
        backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
            in the execution. If ``None``, it uses :class:`qibo.backends.GlobalBackend`.
            Defaults to ``None``.
    """
    nqubits = int(np.log2(dims))

    super_op = random_gaussian_matrix(
        dims**2, rank=rank, mean=0, stddev=1, seed=seed, backend=backend
    )
    super_op = super_op @ np.transpose(np.conj(super_op))

    # partial trace implemented with einsum
    super_op_reduced = np.einsum("ijik->jk", np.reshape(super_op, (dims,) * 4))

    eigenvalues, eigenvectors = np.linalg.eigh(super_op_reduced)

    eigenvalues = np.sqrt(1.0 / eigenvalues)

    operator = np.zeros((dims, dims), dtype=complex)
    operator = backend.cast(operator, dtype=operator.dtype)
    for eigenvalue, eigenvector in zip(eigenvalues, np.transpose(eigenvectors)):
        operator += eigenvalue * np.outer(eigenvector, np.conj(eigenvector))

    if order == "row":
        operator = backend.np.kron(
            backend.identity_density_matrix(nqubits, normalize=False), operator
        )
    if order == "column":
        operator = backend.np.kron(
            operator, backend.identity_density_matrix(nqubits, normalize=False)
        )

    super_op = operator @ super_op @ operator

    return super_op



import numpy as np
import scipy.stats

from qibo import Circuit, gates
from qibo.config import raise_error


class IQAE:
    """Model that performs the Iterative Quantum Amplitude Estimation algorithm.

    The implemented class in this code utilizes the Iterative Quantum Amplitude
    Estimation (IQAE) algorithm, which was proposed in `arxiv:1912.05559
    <https://arxiv.org/abs/1912.05559>`_. The algorithm provides an estimated
    output that, with a probability ``alpha``, differs from the target value by
    ``epsilon``. Both ``alpha`` and ``epsilon`` can be specified.

    Unlike Brassard's original QAE algorithm `arxiv:quant-ph/0005055
    <http://arxiv.org/abs/quant-ph/0005055>`_, this implementation does not rely
    on Quantum Phase Estimation but instead is based solely on Grover's
    algorithm. The IQAE algorithm employs a series of carefully selected Grover
    iterations to determine an estimate for the target amplitude.

    Args:
        circuit_a (:class:`qibo.models.circuit.Circuit`): quantum circuit that
            specifies the QAE problem.

        circuit_q (:class:`qibo.models.circuit.Circuit`): quantum circuit of the
            Grover/Amplification operator.

        alpha (float): confidence level, the target probability is 1 - ``alpha``,
            has values between 0 and 1.

        epsilon (float): target precision for estimation target `a`, has values
            between 0 and 0.5.

        method (str): statistical method used to estimate the confidence
            intervals in each iteration, can be either `chernoff` (default) for
            the Chernoff intervals or `beta` for the Clopper-Pearson intervals.

        n_shots (int): number of shots.

    Raises:
        ValueError: If ``epsilon`` is not in (0, 0.5].
        ValueError: If ``alpha`` is not in (0, 1).
        ValueError: If ``method`` is not supported.
        ValueError: If the number of qubits in ``circuit_a`` is greater than in ``circuit_q``.

    Example:
        .. testcode::

            from qibo import Circuit, gates
            from qibo.models.iqae import IQAE

            # Defining circuit A to integrate sin(x)^2 from [0,1]
            a_circuit = Circuit(2)
            a_circuit.add(gates.H(0))
            a_circuit.add(gates.RY(q = 1, theta = 1 / 2))
            a_circuit.add(gates.CU3(0, 1, 1, 0, 0))
            # Defining circuit Q = -A S_0 A^-1 S_X
            q_circuit = Circuit(2)
            # S_X
            q_circuit.add(gates.Z(q = 1))
            # A^-1
            q_circuit = q_circuit + a_circuit.invert()
            # S_0
            q_circuit.add(gates.X(0))
            q_circuit.add(gates.X(1))
            q_circuit.add(gates.CZ(0, 1))
            # A
            q_circuit = q_circuit + a_circuit

            # Executing IQAE and obtaining the result
            iae = IQAE(a_circuit, q_circuit)
            results = iae.execute()
            integral_value = results.estimation
            integral_error = results.epsilon_estimated
    """

    def __init__(
        self,
        circuit_a,
        circuit_q,
        alpha=0.05,
        epsilon=0.005,
        n_shots=1024,
        method="chernoff",
    ):
        self.circuit_a = circuit_a
        self.circuit_q = circuit_q
        if circuit_a.nqubits > circuit_q.nqubits:
            raise_error(
                ValueError,
                "The number of qubits for Q must be greater or equal than the number"
                "of qubits of A.",
            )
        if not isinstance(n_shots, int):
            raise_error(
                ValueError,
                "The number of shots must be an integer number.",
            )
        # validate ranges of input arguments
        if not 0 < epsilon <= 0.5:
            raise_error(ValueError, f"Epsilon must be in (0, 0.5], but is {epsilon}.")

        if not 0 < alpha < 1:
            raise_error(
                ValueError,
                f"The confidence level alpha must be in (0, 1), but is {alpha}.",
            )

        if method not in {"chernoff", "beta"}:
            raise_error(
                ValueError,
                f"The confidence interval method must be chernoff or beta, but is {method}.",
            )

        self.alpha = alpha
        self.epsilon = epsilon
        self.n_shots = n_shots
        self.method = method

    def construct_qae_circuit(self, k):
        """Generates quantum circuit for QAE.

        Args:
            k (int): number of times the amplification operator ``circuit_q`` is applied.
        Returns:
            The quantum circuit of the QAE algorithm.
        """
        initialization_circuit_a = self.circuit_a
        amplification_circuit_q = self.circuit_q

        qc = Circuit(amplification_circuit_q.nqubits)

        qc.add(
            initialization_circuit_a.on_qubits(
                *range(0, initialization_circuit_a.nqubits, 1)
            )
        )
        for i in range(k):
            qc = qc + amplification_circuit_q
        qc.add(gates.M(initialization_circuit_a.nqubits - 1))
        return qc

    def clopper_pearson(self, count, n, alpha):
        """Calculates the confidence interval for the quantity to estimate `a`.

        Args:
            count (int): number of successes.
            n (int): total number of trials.
            alpha (float): significance level. Must be in (0, 0.5).

        Return:
            The confidence interval [a_min, a_max].
        """
        beta_prob_function = scipy.stats.beta.ppf
        a_min = beta_prob_function(alpha / 2, count, n - count + 1)
        a_max = beta_prob_function(1 - alpha / 2, count + 1, n - count)
        if np.isnan(a_min):
            a_min = 0
        if np.isnan(a_max):
            a_max = 1
        return a_min, a_max

    def h_calc_CP(self, n_successes, n_total_shots, upper_bound_t):
        """Calculates the `h` function.

        Args:
            n_successes (int): number of successes.
            n_total_shots (int): total number of trials.
            upper_bound_t (int): maximum number of rounds to achieve the desired absolute error.

        Returns:
            The h function for the given inputs.
        """
        a_min, a_max = self.clopper_pearson(
            n_successes, n_total_shots, alpha=(self.alpha / upper_bound_t)
        )
        return np.abs(np.arccos(1 - 2 * a_max) - np.arccos(1 - 2 * a_min)) / 2

    def calc_L_range_CP(self, n_shots, upper_bound_t):
        """Calculate the confidence interval for the Clopper-Pearson method.

        Args:
            n_shots (int): number of shots.
            upper_bound_t (int): maximum number of rounds to achieve the desired absolute error.

        Returns:
            max_L, min_L (float, float): The maximum and minimum possible error which could be returned
            on a given iteration.
        """
        x = np.linspace(0, np.pi, 10000)
        x_domain = [x <= 1.0 + 1 / 10 / n_shots]
        y = [
            self.h_calc_CP(int(t * n_shots), n_shots, upper_bound_t)
            for t in x[tuple(x_domain)]
        ]
        max_L = np.max(y) / (2 * np.pi)
        min_L = np.min(y) / (2 * np.pi)
        return max_L, min_L

    def calc_L_range_CH(self, n_shots, upper_bound_t):
        """Calculate the confidence interval for the Chernoff method.

        Args:
            n_shots (int): number of shots.
            upper_bound_t (int): maximum number of rounds to achieve the desired absolute error.

        Returns:
            max_L, min_L (float, float): The maximum and minimum possible error which could be returned
            on a given iteration.
        """
        max_L = (
            np.arcsin(
                (2 / (n_shots) * np.log(2 * upper_bound_t / self.alpha)) ** (1 / 4)
            )
            / 2
            / np.pi
        )
        min_L = np.arcsin(np.sin(max_L) ** 2)
        return max_L, min_L

    def find_next_k(self, uppercase_k_i, up_i, theta_l, theta_u, r=2):
        r"""Find the largest integer ``uppercase_k`` such that the interval ``uppercase_k`` * [ ``theta_l`` , ``theta_u`` ]
        lies completely in [0, `\pi`] or [`\pi`, 2 `\pi`].

        Args:
            uppercase_k_i (int): the current ``uppercase_k`` such ``uppercase_k`` = 4 ``k`` + 2,
                where ``k`` is the power of the operator ``circuit_q``.

            up_i (bool): boolean flag of whether theta_interval lies in the
                upper half-circle [0, `\pi`] or in the lower one [`\pi`, 2 `\pi`].

            theta_l (float): the current lower limit of the confidence interval for the angle theta.

            theta_u (float): the current upper limit of the confidence interval for the angle theta.

            r (int): lower bound for ``uppercase_k``.

        Returns:
            The next power `K_i`, and boolean flag for the extrapolated interval.
        """
        uppercase_k_max = int(1 / (2 * (theta_u - theta_l)))
        uppercase_k = uppercase_k_max - (uppercase_k_max - 2) % 4
        while uppercase_k >= r * uppercase_k_i:
            theta_min = uppercase_k * theta_l - int(uppercase_k * theta_l)
            theta_max = uppercase_k * theta_u - int(uppercase_k * theta_u)
            if int(uppercase_k * theta_u) == int(uppercase_k * theta_l):
                if theta_max <= 1 / 2 and theta_min <= 1 / 2:
                    up = True
                    return (uppercase_k, up)
                elif theta_max >= 1 / 2 and theta_min >= 1 / 2:
                    up = False
                    return (uppercase_k, up)
            uppercase_k -= 4
        return (uppercase_k_i, up_i)

    def execute(self, backend=None):
        """Execute IQAE algorithm.

        Args:
            backend: the qibo backend.

        Returns:
            A :class:`qibo.models.iqae.IterativeAmplitudeEstimationResult` results object.
        """
        from qibo.backends import _check_backend

        backend = _check_backend(backend)

        # Initializing all parameters
        k = [0]
        # uppercase_k=4k+2
        uppercase_k = [2]

        theta_u = 1 / 4
        theta_l = 0
        theta_intervals = [theta_l, theta_u]
        a_intervals = [
            np.sin(2 * np.pi * theta_l) ** 2,
            np.sin(2 * np.pi * theta_u) ** 2,
        ]
        a_min = [0]
        a_max = [1]
        theta_dif = np.abs(theta_u - theta_l)
        up = [True]
        samples_history = []
        n_shots_history = []

        eps = self.epsilon / (2 * np.pi)
        upper_bound_t = int(np.log2(np.pi / (8 * self.epsilon))) + 1
        n_total_shots = self.n_shots
        num_oracle_queries = 0

        if self.method == "chernoff":
            # Chernoff method
            max_L, min_L = self.calc_L_range_CH(n_total_shots, upper_bound_t)

        else:
            # Clopper-Pearson (beta) method
            max_L, min_L = self.calc_L_range_CP(n_total_shots, upper_bound_t)

        i = 0
        while theta_dif > 2 * eps:
            i = i + 1
            uppercase_k_i, up_i = self.find_next_k(
                uppercase_k[-1], up[-1], theta_l, theta_u
            )
            k_i = int((uppercase_k_i - 2) / 4)
            uppercase_k.append(uppercase_k_i)
            up.append(up_i)
            k.append(k_i)
            # Checking the no-overshooting condition
            if uppercase_k_i > int(max_L / eps):
                # We ensure to not make unnecessary measurement shots at last iterations of the algorithm
                n_shots_i = int((max_L / eps) * self.n_shots / uppercase_k_i / 10)
                # To avoid having a null number of shots
                if n_shots_i == 0:
                    n_shots_i = 1
            else:
                n_shots_i = self.n_shots

            # Calling and executing the quantum circuit
            qc = self.construct_qae_circuit(k_i)
            samples = qc(nshots=n_shots_i).frequencies(binary=True)["1"]

            samples_history.append(samples)
            n_shots_history.append(n_shots_i)

            num_oracle_queries += n_shots_i * k_i

            m = 1
            if i > 1:
                while uppercase_k[i - m] == uppercase_k[i] and i >= m + 1:
                    m += 1
                sum_total_samples = sum([samples_history[j] for j in range(i - m, i)])
                n_total_shots = sum([n_shots_history[j] for j in range(i - m, i)])

            else:
                sum_total_samples = samples
                n_total_shots = n_shots_i

            a = sum_total_samples / n_total_shots

            if self.method == "chernoff":
                delta_a = np.sqrt(
                    np.log(2 * upper_bound_t / self.alpha) / 2 / n_total_shots
                )
                a_min_i = max(0, a - delta_a)
                a_max_i = min(1, a + delta_a)
            else:
                a_min_i, a_max_i = self.clopper_pearson(
                    sum_total_samples, n_total_shots, alpha=self.alpha / upper_bound_t
                )
            a_min.append(a_min_i)
            a_max.append(a_max_i)

            if up_i:
                theta_min_i = np.arccos(1 - 2 * a_min_i) / 2 / np.pi
                theta_max_i = np.arccos(1 - 2 * a_max_i) / 2 / np.pi
            else:
                theta_min_i = 1 - np.arccos(1 - 2 * a_max_i) / 2 / np.pi
                theta_max_i = 1 - np.arccos(1 - 2 * a_min_i) / 2 / np.pi

            theta = min(int(uppercase_k_i * theta_u), int(uppercase_k_i * theta_l))

            theta_u = (theta + theta_max_i) / uppercase_k_i
            theta_l = (theta + theta_min_i) / uppercase_k_i
            theta_dif = np.abs(theta_u - theta_l)
            theta_intervals.append([theta_l, theta_u])

            a_l = np.sin(2 * np.pi * theta_l) ** 2
            a_u = np.sin(2 * np.pi * theta_u) ** 2
            a_intervals.append([a_l, a_u])

        result = IterativeAmplitudeEstimationResult()
        result.alpha = self.alpha
        result.epsilon_target = self.epsilon
        result.epsilon_estimated = (a_u - a_l) / 2
        result.estimate_intervals = a_intervals
        result.num_oracle_queries = num_oracle_queries
        result.estimation = (a_u + a_l) / 2
        result.theta_intervals = theta_intervals
        result.k_list = k
        result.ratios = samples_history
        result.shots = n_shots_history

        return result


class IterativeAmplitudeEstimationResult:
    """The ``IterativeAmplitudeEstimationResult`` result object."""

    def __init__(self):
        self._alpha = None
        self._epsilon_target = None
        self._epsilon_estimated = None
        self._num_oracle_queries = None
        self._estimation = None
        self._estimate_intervals = None
        self._theta_intervals = None
        self._k_list = None
        self._ratios = None
        self._shots = None



import numpy as np

from qibo import gates
from qibo.config import log, raise_error
from qibo.models.circuit import Circuit


class Grover:
    """Model that performs Grover's algorithm.

    For Grover's original search algorithm: `arXiv:quant-ph/9605043 <https://arxiv.org/abs/quant-ph/9605043>`_
    For the iterative version with unknown solutions:`arXiv:quant-ph/9605034 <https://arxiv.org/abs/quant-ph/9605034>`_
    For the Grover algorithm with any superposition:`arXiv:quant-ph/9712011 <https://arxiv.org/abs/quant-ph/9712011>`_

    Args:
        oracle (:class:`qibo.core.circuit.Circuit`): quantum circuit that flips
            the sign using a Grover ancilla initialized with -X-H-. Grover ancilla
            expected to be last qubit of oracle circuit.
        superposition_circuit (:class:`qibo.core.circuit.Circuit`): quantum circuit that
            takes an initial state to a superposition. Expected to use the first
            set of qubits to store the relevant superposition.
        initial_state_circuit (:class:`qibo.core.circuit.Circuit`): quantum circuit
            that initializes the state. If empty defaults to ``|000..00>``
        superposition_qubits (int): number of qubits that store the relevant superposition.
            Leave empty if superposition does not use ancillas.
        superposition_size (int): how many states are in a superposition.
            Leave empty if its an equal superposition of quantum states.
        number_solutions (int): number of expected solutions. Needed for normal Grover.
            Leave empty for iterative version.
        target_amplitude (float): absolute value of the amplitude of the target state. Only for
            advanced use and known systems.
        check (function): function that returns True if the solution has been
            found. Required of iterative approach.
            First argument should be the bitstring to check.
        check_args (tuple): arguments needed for the check function.
            The found bitstring not included.
        iterative (bool): force the use of the iterative Grover

    Example:
        .. testcode::

            import numpy as np

            from qibo import Circuit, gates
            from qibo.models.grover import Grover

            # Create an oracle. Ex: Oracle that detects state |11111>
            oracle = Circuit(5 + 1)
            oracle.add(gates.X(5).controlled_by(*range(5)))

            # Create superoposition circuit. Ex: Full superposition over 5 qubits.
            superposition = Circuit(5)
            superposition.add([gates.H(i) for i in range(5)])

            # Generate and execute Grover class
            grover = Grover(oracle, superposition_circuit=superposition, number_solutions=1)
            solution, iterations = grover()
    """

    def __init__(
        self,
        oracle,
        superposition_circuit=None,
        initial_state_circuit=None,
        superposition_qubits=None,
        superposition_size=None,
        number_solutions=None,
        target_amplitude=None,
        check=None,
        check_args=(),
        iterative=False,
    ):
        self.oracle = oracle
        self.initial_state_circuit = initial_state_circuit

        if superposition_circuit:
            self.superposition = superposition_circuit
        else:
            if not superposition_qubits:
                raise_error(
                    ValueError,
                    "Cannot create Grover model if the "
                    "superposition circuit or number of "
                    "qubits is not specified.",
                )
            self.superposition = Circuit(superposition_qubits)
            self.superposition.add([gates.H(i) for i in range(superposition_qubits)])

        if superposition_qubits:
            self.sup_qubits = superposition_qubits
        else:
            self.sup_qubits = self.superposition.nqubits

        if superposition_size:
            self.sup_size = superposition_size
        else:
            self.sup_size = int(2**self.sup_qubits)

        assert oracle.nqubits > self.sup_qubits

        self.anc_qubits_sup = self.superposition.nqubits - self.sup_qubits
        self.anc_qubits_ora = self.oracle.nqubits - self.sup_qubits - 1

        self.nqubits = (
            self.sup_qubits + max(self.anc_qubits_sup, self.anc_qubits_ora) + 1
        )

        self.check = check
        self.check_args = check_args
        self.num_sol = number_solutions
        self.targ_a = target_amplitude
        self.iterative = iterative

        self.space_sup = list(range(self.sup_qubits + self.anc_qubits_sup))
        self.space_ora = list(range(self.sup_qubits + self.anc_qubits_ora)) + [
            self.nqubits - 1
        ]

    def initialize(self):
        """Initialize the Grover algorithm with the superposition and Grover ancilla."""
        c = Circuit(self.nqubits)
        c.add(gates.X(self.nqubits - 1))
        c.add(gates.H(self.nqubits - 1))
        if self.initial_state_circuit:
            c.add(
                self.initial_state_circuit.invert().on_qubits(
                    *range(self.initial_state_circuit.nqubits)
                )
            )
        c.add(self.superposition.on_qubits(*self.space_sup))
        return c

    def diffusion(self):
        """Construct the diffusion operator out of the superposition circuit."""
        nqubits = self.superposition.nqubits + 1
        c = Circuit(nqubits)
        c.add(self.superposition.invert().on_qubits(*range(nqubits - 1)))
        if self.initial_state_circuit:
            c.add(
                self.initial_state_circuit.invert().on_qubits(
                    *range(self.initial_state_circuit.nqubits)
                )
            )
        c.add([gates.X(i) for i in range(self.sup_qubits)])
        c.add(gates.X(nqubits - 1).controlled_by(*range(self.sup_qubits)))
        c.add([gates.X(i) for i in range(self.sup_qubits)])
        if self.initial_state_circuit:
            c.add(
                self.initial_state_circuit.on_qubits(
                    *range(self.initial_state_circuit.nqubits)
                )
            )
        c.add(self.superposition.on_qubits(*range(nqubits - 1)))
        return c

    def step(self):
        """Combine oracle and diffusion for a Grover step."""
        c = Circuit(self.nqubits)
        c.add(self.oracle.on_qubits(*self.space_ora))
        c.add(self.diffusion().on_qubits(*(self.space_sup + [self.nqubits - 1])))
        return c

    def circuit(self, iterations):
        """Creates circuit that performs Grover's algorithm with a set amount of iterations.

        Args:
            iterations (int): number of times to repeat the Grover step.

        Returns:
            :class:`qibo.core.circuit.Circuit` that performs Grover's algorithm.
        """
        c = Circuit(self.nqubits)
        c += self.initialize()
        for _ in range(iterations):
            c += self.step()
        c.add(gates.M(*range(self.sup_qubits)))
        return c

    def iterative_grover(self, lamda_value=6 / 5, backend=None):
        """Iterative approach of Grover for when the number of solutions is not known.

        Args:
            lamda_value (real): parameter that controls the evolution of the iterative method.
                                Must be between 1 and 4/3.
            backend (:class:`qibo.backends.abstract.Backend`): Backend to use for circuit execution.

        Returns:
            measured (str): bitstring measured and checked as a valid solution.
            total_iterations (int): number of times the oracle has been called.
        """
        from qibo.backends import _check_backend

        backend = _check_backend(backend)

        k = 1
        lamda = lamda_value
        total_iterations = 0
        while True:
            it = np.random.randint(k + 1)
            if it != 0:
                total_iterations += it
                circuit = self.circuit(it)
                result = backend.execute_circuit(circuit, nshots=1)
                measured = result.frequencies(binary=True).most_common(1)[0][0]
                if self.check(measured, *self.check_args):
                    return measured, total_iterations
            k = min(lamda * k, np.sqrt(self.sup_size))
            if total_iterations > (9 / 4) * np.sqrt(self.sup_size):
                log.warning("Too many total iterations, output might not be solution.")
                return measured, total_iterations

    def execute(self, nshots=100, freq=False, logs=False, backend=None):
        """Execute Grover's algorithm.

        If the number of solutions is given, calculates iterations,
        otherwise it uses an iterative approach.

        Args:
            nshots (int): number of shots in order to get the frequencies.
            freq (bool): print the full frequencies after the exact Grover algorithm.
            backend (:class:`qibo.backends.abstract.Backend`): Backend to use for circuit execution.

        Returns:
            solution (str): bitstring (or list of bitstrings) measured as solution of the search.
            iterations (int): number of oracle calls done to reach a solution.
        """
        from qibo.backends import _check_backend

        backend = _check_backend(backend)

        if (self.num_sol or self.targ_a) and not self.iterative:
            if self.targ_a:
                it = int(np.pi * (1 / self.targ_a) / 4)
            else:
                it = int(np.pi * np.sqrt(self.sup_size / self.num_sol) / 4)
            circuit = self.circuit(it)
            result = backend.execute_circuit(circuit, nshots=nshots)
            result = result.frequencies(binary=True)
            if freq:
                if logs:
                    log.info("Result of sampling Grover's algorihm")
                    log.info(result)
                self.frequencies = result
            if logs:
                log.info(
                    f"Most common states found using Grover's algorithm with {it} iterations:"
                )
            if self.targ_a:
                most_common = result.most_common(1)
            else:
                most_common = result.most_common(self.num_sol)
            self.solution = []
            self.iterations = it
            for i in most_common:
                if logs:
                    log.info(i[0])
                self.solution.append(i[0])
                if logs:
                    if self.check:
                        if self.check(i[0], *self.check_args):
                            log.info("Solution checked and successful.")
                        else:
                            log.info(
                                "Not a solution of the problem. Something went wrong."
                            )
        else:
            if not self.check:
                raise_error(ValueError, "Check function needed for iterative approach.")
            measured, total_iterations = self.iterative_grover(backend=backend)
            if logs:
                log.info("Solution found in an iterative process.")
                log.info(f"Solution: {measured}")
                log.info(f"Total Grover iterations taken: {total_iterations}")
            self.solution = measured
            self.iterations = total_iterations
        return self.solution, self.iterations

    def __call__(self, nshots=100, freq=False, logs=False, backend=None):
        """Equivalent to :meth:`qibo.models.grover.Grover.execute`."""
        return self.execute(nshots=nshots, freq=freq, logs=logs, backend=backend)
